/*****************************************************************************
 *
 * PROJECT: Task Control Management.
 *
 * (c) Copyright 1996 Reid Simmons.  All rights reserved.
 *
 * FILE: tcmHandle.h
 *
 * ABSTRACT: Basic definitions needed by TCM
 *
 * $Source: /afs/cs/project/TCA/Master/tcaV8/tcm/src/tcmHandle.h,v $ 
 * $Revision: 1.4 $
 * $Date: 1999/02/14 01:11:50 $
 * $Author: da0g $
 *
 * REVISION HISTORY:
 *
 * $Log: tcmHandle.h,v $
 * Revision 1.4  1999/02/14 01:11:50  da0g
 * Updated to work with g++ 2.8.1 and egcs.
 *
 * Revision 1.3  97/12/29  17:06:35  reids
 * Version that has the basic functionality needed to support TDL.
 * 
 * Revision 1.2  97/12/04  17:50:27  reids
 * Another fairly stable version (except that monitors do not quite work)
 * 
 * Revision 1.1  97/11/21  14:06:45  reids
 * First release of TCM -- seems to be a stable version
 * 
 *
 *****************************************************************************/

#ifndef INCtcmHandle
#define INCtcmHandle

class Ref_Count
{
 public:
  Ref_Count() { refCount = 0; }
  virtual ~Ref_Count() { }

  void increment(void) { refCount++; }
  void decrement(void) { refCount--; if (refCount == 0) { delete this; } }

 private:
  int refCount;
};

#ifndef macintosh
template <class T> void refIncr(T *ptr);// { ptr->increment(); }
template <class T> void refDecr(T *ptr);// { ptr->decrement(); }

  /* Note:  The new [1998] ANSI C++ Standard has changed the way templated
   * function specializations are declared.  Both g++ 2.8.1 and egcs support
   * the "new" Standard.  Which, unfortunately, is syntactically incompatible
   * with the "old" Standard supported by g++ 2.7.2.x.
   */
#if    defined ( ANSI_STANDARD_CXX ) \
    || ( __GNUC__ == 2 && ( __GNUC_MINOR__ == 90 || __GNUC_MINOR__ == 8 ) )

#define INSTANTIATE_REF_COUNT_FUNCTIONS(type) \
void refIncr(type *ptr) { ptr->increment(); } \
void refDecr(type *ptr) { ptr->decrement(); } \
template <> void refIncr<type>(type *ptr) { ptr->increment(); } \
template <> void refDecr<type>(type *ptr) { ptr->decrement(); }

#else

#define INSTANTIATE_REF_COUNT_FUNCTIONS(type) \
void refIncr(type *ptr) { ptr->increment(); } \
void refDecr(type *ptr) { ptr->decrement(); }

#endif

#endif

template <class T>
class tcmHandle
{
 public:
  tcmHandle() { refPtr = NULL; }
  tcmHandle(T *ptr) { _bind(ptr); }
  tcmHandle(tcmHandle<T> const &handle) { _bind(handle.refPtr); }
  ~tcmHandle() { _release(); }

  tcmHandle<T> &operator= (tcmHandle<T> const &handle) 
    { _release(); _bind(handle.refPtr); return *this; }
  tcmHandle<T> &operator= (T *ptr) { _release(); _bind(ptr); return *this; }
  bool operator== (tcmHandle<T> const &handle) const
    { return refPtr == handle.refPtr; }
  // Silly, but some classes need it defined
  bool operator< (tcmHandle<T> const &handle) const 
    { return (int)refPtr < (int)handle.refPtr; }

  T * operator* () const { return refPtr; }
  T *operator-> () const { return refPtr; }

 private:
  T * refPtr;

#ifdef macintosh
  void _release (void) { if (refPtr) { ((Ref_Count *)refPtr)->decrement(); refPtr = NULL; } }
  void _bind (T *ptr) { refPtr = ptr; if (ptr) { ((Ref_Count *)ptr)->increment(); } }
#else
  void _release (void) { if (refPtr) { refDecr(refPtr); refPtr = NULL; } }
  void _bind (T *ptr) { refPtr = ptr; if (ptr) { refIncr(ptr); } }
#endif
};

template <class T>
inline tcmHandle<T> makeHandle(T *ptr)
{
  tcmHandle<T> handle(ptr);
  return handle;
}

#endif /* INChandle */
