
public interface DataComponentObserver
{
	/* There is a fairly complex data structure here.  There is an
	 * observable object that is being watched, and an observer object
	 * that is doing the watching.  There may also be an additional
	 * third object that is the parent of the observable object, and
	 * which will have its own set of observers (watchers).
	 *
	 * This interface (the update method) is invoked when an observable
	 * object changes to notify its observers.  At this time, the
	 * "theObservedComponent" is the component that is being watched,
	 * and the "theModifiedComponent" is the component that has
	 * actually changed.  These two values may be identical.  Or
	 * "theModifiedComponent" may be a child (or grandchild, et al) of
	 * "theObservedComponent".
	 *
	 * "theStoredData" is always whatever data the observer (watching)
	 * object has chosen to store with the observable (watched) object
	 * for identification or convenience.
	 */
  public boolean update ( DataComponent  theObservedComponent,
			  DataComponent  theModifiedComponent,
			  Object         theStoredData        );

	/** Convenience method for unregistering DataComponentObserver's */
  public boolean unregisterToObserve ();
}
