
import netscape.application.Color;

/*
 * This represents a C "do {} while (...);" statement.
 */

public class DataDoWhileStatement extends DataIterationStatement
{
	/* Class variables */
  public final static String   DO               = "do";
  public final static String   WHILE            = "while";
  public final static String   OPEN_PAREN       = "(";
  public final static String   CLOSE_PAREN      = ")";
  public final static String   EXPRESSION_INDEX = "ExpressionIndex";
  public final static String   STATEMENT_INDEX  = "StatementIndex";
  public final static String   SEMICOLON        = ";";

  public       static Color    DO_WHILE_COLOR           = Color.magenta;
  public final static int      DO_WHILE_EXPRESSION_ONLY = 7;


	/* Instance Variables */
  protected DataExpression           expression;


	/* Instance Methods */
  public DataDoWhileStatement()
  {
    expression = null;
  }

  public DataExpression getExpression () { return expression; }

  public void           setExpression ( DataExpression theExpression )
  {
    if ( expression != null )
      expression . setParent ( null );

    expression = theExpression;

    if ( expression != null )
      expression . setParent ( this );
  }


	/* Allow searching of our expression... */
  public boolean hasSubcomponentFraction ( String theString )
  {
    return super               . hasSubcomponentFraction ( theString )
      || (   ( getExpression() != null )
	  && ( getExpression() . hasSubcomponentFraction ( theString ) ) );
  }


  public DataComponent newComponent ( )
  {
    DataDoWhileStatement  doWhileStatement = new DataDoWhileStatement();

    doWhileStatement . addSubcomponent ( "\n" );
    doWhileStatement . setIndex ( DataDoWhileStatement.DO );
    doWhileStatement . setIndex ( DataDoWhileStatement.WHILE );
    doWhileStatement . addSubcomponent ( " " );
    doWhileStatement . setIndex ( DataDoWhileStatement.OPEN_PAREN  );
    doWhileStatement . setIndex ( DataDoWhileStatement.CLOSE_PAREN );
    doWhileStatement . addSubcomponent ( " " );
    doWhileStatement . setIndex ( DataDoWhileStatement.SEMICOLON );

    return doWhileStatement;
  }


  public String getWarnString ( int      theObjectSubset,
				boolean  theIsGenerating )
  {
    return super . getWarnString ( theObjectSubset, theIsGenerating )
      + " or DataDoWhileStatement.DO_WHILE_EXPRESSION_ONLY ("
      + DataDoWhileStatement.DO_WHILE_EXPRESSION_ONLY + ")";
  }

  public boolean isValidObjectSubset ( int      theObjectSubset,
				       boolean  theIsGenerating )
  {
    if ( theObjectSubset == DataDoWhileStatement.DO_WHILE_EXPRESSION_ONLY )
      return true;
    else
      return super . isValidObjectSubset ( theObjectSubset, theIsGenerating );
  }


  public boolean isValid ( int theObjectSubsetToValidate )
  {
      /* We should check to make sure it parses later on. *DBA*/

    if (    theObjectSubsetToValidate
	 == DataDoWhileStatement.DO_WHILE_EXPRESSION_ONLY )
    {
      return getExpression() != null;
    }

	/* Simple checks that we can do here.. */
    if (   ( getExpression() == null )
	|| ( getStatement()  == null ) )
      return false;

    return super.isValid ( theObjectSubsetToValidate );
  }


  protected
    DataComponent parseSpecificComponent ( TDLParser theTDLParser,
					   int       theObjectSubsetToParse )
    throws Exception
  {
    warnIfInvalidObjectSubset ( theObjectSubsetToParse,
				"parseSpecificComponent", false );
    
    if (    theObjectSubsetToParse
	 == DataDoWhileStatement.DO_WHILE_EXPRESSION_ONLY )
      return theTDLParser . parseExpressionObject ( 0 );
    else
      return theTDLParser . parseDoWhile ( 0 );
  }


  public boolean setFrom ( DataComponent  theObjectToCopy,
			   int            theObjectSubsetToSet,
			   boolean        theCanDestroyObject  )
  {
    DataComponent  aObject = theObjectToCopy;

    if (   (    theObjectSubsetToSet
	     == DataDoWhileStatement.DO_WHILE_EXPRESSION_ONLY )
	&& ( theObjectToCopy instanceof DataExpression     ) )
    {
      if (    ( theCanDestroyObject == false )
	  &&  ( aObject             != null  ) )
      {
	aObject = aObject . copyComponent();
	if ( aObject == null )
	{
	  System.err.println ( "[DataDoWhileStatement:setFrom]  Error:  "
		       + "Unable to replicate object:  " + theObjectToCopy );
	  return false;
	}
      }
      setExpression ( (DataExpression) aObject );
      return true;
    }


    if ( /*NOT*/ !  ( theObjectToCopy instanceof DataDoWhileStatement ) )
    {
      System.err.println ( "[DataDoWhileStatement:setFrom]  Error:  "
			   + "Illegal object.  (\""
			   + theObjectToCopy . getClass() . getName()
			   + "\")" );
      return false;
    }

    DataDoWhileStatement  theDoWhileStatementToCopy
			      = (DataDoWhileStatement) theObjectToCopy;


    aObject = theDoWhileStatementToCopy.expression;

    if (    ( theCanDestroyObject == false )
	&&  ( aObject             != null  ) )
    {
      aObject = aObject . copyComponent();
      if ( aObject == null )
      {
	System.err.println ( "[DataDoWhileStatement:setFrom]  ERROR:  "
			     + "Unable to replicate object:  "
			     + theDoWhileStatementToCopy.expression);
	return false;
      }
    }
    setExpression ( (DataExpression) aObject );


    if (    theObjectSubsetToSet
	 == DataDoWhileStatement.DO_WHILE_EXPRESSION_ONLY )
    {
      return true;
    }
    
    return super . setFrom ( theObjectToCopy, theObjectSubsetToSet,
			     theCanDestroyObject );
  }


  public void generateExpression ( DataDestination theOutputDestination,
				   int             theObjectSubsetToGenerate,
				   boolean         theIsPartOfDoWhileStatement)
  {
    boolean  needsLeadingSpace  = theIsPartOfDoWhileStatement,
             needsTrailingSpace = theIsPartOfDoWhileStatement;

    if ( getExpression() != null )
    {
      if ( getExpression() . hasLeadingWhitespaceSubcomponent() )
	needsLeadingSpace = false;
      if ( getExpression() . hasTrailingWhitespaceSubcomponent() )
	needsTrailingSpace = false;
    }

    	/* Initialize us to generate non-significant tokens... */
    initializeGenerateSubcomponentIndex ( DataDoWhileStatement.OPEN_PAREN );

	/* Write any pre-EXPRESSION non-significant tokens */
    generateSubcomponents ( DataDoWhileStatement.EXPRESSION_INDEX,
			    theOutputDestination,
			    theObjectSubsetToGenerate,
			    needsLeadingSpace );

	/* Write out expression, if we have one... */
    if ( getExpression() != null )
      getExpression() . generate ( theOutputDestination,
				   theObjectSubsetToGenerate );


	/* Write any pre-")" tokens */
    generateSubcomponents ( DataDoWhileStatement.CLOSE_PAREN,
			    theOutputDestination,
			    theObjectSubsetToGenerate,
			    needsTrailingSpace );
  }


  public void generate ( DataDestination  theOutputDestination,
			 int              theObjectSubsetToGenerate )
  {
    int i;

       	/* Print a warning if necessary... */
    warnIfInvalidObjectSubset ( theObjectSubsetToGenerate, "generate", true );

		/* Handle subset case... */
    if (    theObjectSubsetToGenerate
	 == DataDoWhileStatement.DO_WHILE_EXPRESSION_ONLY )
    {
      generateExpression ( theOutputDestination, DataComponent.ENTIRE_OBJECT,
			   false );
      return;
    }

    	/* Initialize us to generate non-significant tokens... */
    initializeGenerateSubcomponentIndex();

	/* Generate any labels that we have... */
    generateLabels ( theOutputDestination,
		     theObjectSubsetToGenerate,
		     DataDoWhileStatement.DO,
		     getLabel() == null /* Disable #line macros if no labels.*/
		    );

	/* Write any pre-do non-significant tokens */
    generateSubcomponents ( DataDoWhileStatement.DO,
			    theOutputDestination,
			    theObjectSubsetToGenerate, false );

	/* Generate extra pre-iteration code. */
    generatePreIterationCode ( theOutputDestination,
			       theObjectSubsetToGenerate );

	/* Generate relevant #line-number macros */
    generateRelevantLineNumberMacros ( theOutputDestination,
				       getLineNumber(),
				       0 );

	/* Write our DO keyword */
    theOutputDestination . write ( DataDoWhileStatement.DO );



	/* Write any pre-STATEMENT non-significant tokens */
    generateSubcomponents ( DataDoWhileStatement.STATEMENT_INDEX,
			    theOutputDestination,
			    theObjectSubsetToGenerate, false );


	/* Write the statement, if we have one... */
    if ( getStatement() != null )
    {
	/* Generate extra inside-iteration code */
      addInsideIterationCode ( theObjectSubsetToGenerate, getStatement() );

      getStatement() . generate ( theOutputDestination,
				  theObjectSubsetToGenerate );

	/* Remove extra inside-iteration code */
      removeInsideIterationCode ( theObjectSubsetToGenerate, getStatement() );
    }
    else  /* Try to recover gracefully if we have a problem... */
    {
	/* Generate empty body statement with extra inside-iteration code */
      generateEmptyIterationBodyStatement ( theOutputDestination,
					    theObjectSubsetToGenerate );
    }


	/* Write any pre-while non-significant tokens */
    generateSubcomponents ( DataDoWhileStatement.WHILE,
			    theOutputDestination,
			    theObjectSubsetToGenerate, true );

	/* Write our WHILE keyword */
    theOutputDestination . write ( DataDoWhileStatement.WHILE );


	/* Write any pre-"(" non-significant tokens */
    generateSubcomponents ( DataDoWhileStatement.OPEN_PAREN,
			    theOutputDestination,
			    theObjectSubsetToGenerate, true );

	/* Write "(" */
    theOutputDestination . write ( DataDoWhileStatement.OPEN_PAREN );


	/* Write pre-Expression tokens, the expression, and pre-")" tokens */
    generateExpression ( theOutputDestination, theObjectSubsetToGenerate,
			 true );


	/* Write ")" */
    theOutputDestination . write ( DataDoWhileStatement.CLOSE_PAREN );


	/* Write any pre-";" tokens */
    generateSubcomponents ( DataDoWhileStatement.SEMICOLON,
			    theOutputDestination,
			    theObjectSubsetToGenerate, false );

	/* Write ";" */
    theOutputDestination . write ( DataDoWhileStatement.SEMICOLON );



	/* Generate extra post-iteration code. */
    generatePostIterationCode ( theOutputDestination,
				theObjectSubsetToGenerate );

	/* Write any remaining non-significant tokens */
	/* (There *SHOULD* *NOT* be any...  But just in case...) */
    generateAllRemainingSubcomponents ( theOutputDestination,
					theObjectSubsetToGenerate, false );
  }



	/********************************************************/
        /********** Methods to interface with Graphing **********/
	/********************************************************/
  public boolean supportsGraphType ( int theGraphType )
  {
    if (   ( theGraphType == GraphRoot.GUI_TREE_GRAPH ) )
      return true;
    else
      return false;
  }

  public GraphModule createStatementGraphModule ( int theGraphType )
  {
    if (   ( theGraphType != GraphRoot.GUI_TREE_GRAPH ) )
      return null;

    /* else ... */

    /* Building:  +- DO - +
     *            |  |    |
     *            |  |    
     *            |  while - "[.]"
     */

    ScrollTextOfDataComponent  expressionText
      = new ScrollTextOfDataComponent ( this, DataDoWhileStatement
					       . DO_WHILE_EXPRESSION_ONLY );

    GraphNode       expressionNode = new GraphNode ( expressionText,
					   GraphComponent.LEFT,
					   DataDoWhileStatement.DO_WHILE_COLOR,
					   true, null );
						     
    GraphConnection toExpression = new GraphConnection ( GraphComponent.LEFT,
					   DataDoWhileStatement.DO_WHILE_COLOR,
					   false, false, expressionNode );

    GraphNode       whileNode = new GraphNode ( new SimpleLabel( "WHILE" ),
					   GraphComponent.UP,
					   DataDoWhileStatement.DO_WHILE_COLOR,
					   false, toExpression );

    GraphConnection toWhile = new GraphConnection ( GraphComponent.UP,
					   DataDoWhileStatement.DO_WHILE_COLOR,
					   false, true, whileNode );


    GraphConnection toChildren = new GraphConnection ( GraphComponent.UP,
					   DataComponent.getDraggableColor(),
					   true, true, null );

    GraphMeet       childrenMeet = new GraphMeet ( GraphComponent.LEFT,
					   DataDoWhileStatement.DO_WHILE_COLOR,
					   GraphComponent.DOWN, toChildren );

    GraphConnection toChildrenMeet = new GraphConnection ( GraphComponent.LEFT,
					   DataDoWhileStatement.DO_WHILE_COLOR,
					   false, false, childrenMeet );

    GraphNode       doNode = new GraphNode ( new SimpleLabel ( "DO" ),
				           GraphComponent.LEFT,
					   DataDoWhileStatement.DO_WHILE_COLOR,
					   false, toChildrenMeet );
    doNode . addChild ( toWhile );



    GraphConnection toDo = new GraphConnection ( GraphComponent.LEFT,
					   DataDoWhileStatement.DO_WHILE_COLOR,
					   false, false, doNode );

    GraphMeet       startMeet = new GraphMeet ( GraphComponent.UP,
					   DataDoWhileStatement.DO_WHILE_COLOR,
					   GraphComponent.LEFT, toDo );

    GraphConnection tail = new GraphConnection ( GraphComponent.UP,
					   DataComponent.getDraggableColor(),
					   true, true, null );
    startMeet . addChild ( tail );

    GraphModule  graphModule = new GraphModule ( startMeet, tail, toChildren,
						 null, this, true,
						 theGraphType );

    updateGraphModule ( this, null, graphModule );

    return graphModule;
  }



	/* Invoked by this.update(*) invoking GraphModule.update() */
  public boolean updateGraphModule ( DataComponent  theModifiedComponent,
				     Object         theStoredData,
				     GraphModule    theGraphModule )
  {
	/* Skip updates to parents... */
    if ( isPartOfThisStatement ( theModifiedComponent ) == false )
      return true;

    if ( getStatement() != null )
      return getStatement() . updateGraphModule( theGraphModule,
						 GraphModule.CHILD_COMPONENT );
    else
      return true;
  }

}
