
import netscape.util.*;

public class DataTaskArgument extends DataComponent
{
	/* Class Constants */
  public    final static int       IN_DIRECTION      = 0;
  public    final static int       OUT_DIRECTION     = 1;
  public    final static int       INOUT_DIRECTION   = 2;
  public    final static int       NO_DIRECTION      = 3;
  public    final static int       DEFAULT_DIRECTION = NO_DIRECTION;
  public    final static int       INVALID_DIRECTION = -1;
  protected final static String[]  ARGUMENT_TYPES = { "IN", "OUT", "INOUT",""};

  public    final static String    PERSISTENT               = "PERSISTENT";

  public    final static String    ARGUMENT_DIRECTION_INDEX = "ArgDir";
  public    final static String    ARGUMENT_NAME_INDEX      = "ArgName";
	/* Note:  Index *OF* the equals-sign */
  public    final static String    ARGUMENT_EQUALS_INDEX    = "ArgEquals";

	/* Possible subsets of this task-argument. */
  public    final static int       TYPE_NAME_AND_EQUALS = 201;
  public    final static int       TYPE_AND_NAME        = 202;
  public    final static int       AFTER_EQUALS         = 203;


	/* Class methods */
  public static int getArgumentDirectionOfString ( String theArgumentDirection)
  {
    for ( int i=0;  i < DataTaskArgument.ARGUMENT_TYPES.length;  i++ )
    {
      if ( DataTaskArgument.ARGUMENT_TYPES [ i ]
	   . equalsIgnoreCase ( theArgumentDirection ) )
      {
	return i;
      }
    }
    return DataTaskArgument . INVALID_DIRECTION;
  }


	/* Instance Data */
  protected boolean isPersistentDeclaration;
  protected int     argumentDirection;
  protected String  argumentName;
  protected Vector  tmpVector;  /* Temporary object used in setFrom() */


	/* Instance Methods */
  public DataTaskArgument ( )
  {
    isPersistentDeclaration = false;
    tmpVector = new Vector();
    setArgumentDirection ( DataTaskArgument . DEFAULT_DIRECTION );
    setArgumentNameWithoutParsingChecks ( DataComponent . EMPTY_STRING );
  }

  public boolean getIsPersistentDeclaration()
					    { return isPersistentDeclaration; }
  public void    setIsPersistentDeclaration(
					  boolean theIsPersistentDeclaration )
		      { isPersistentDeclaration = theIsPersistentDeclaration; }


  public int  getArgumentDirection ( ) { return argumentDirection; }
  public void setArgumentDirection ( int theArgumentDirection )
  {
    if (   ( theArgumentDirection != DataTaskArgument . IN_DIRECTION    )
	&& ( theArgumentDirection != DataTaskArgument . OUT_DIRECTION   )
	&& ( theArgumentDirection != DataTaskArgument . INOUT_DIRECTION )
	&& ( theArgumentDirection != DataTaskArgument . NO_DIRECTION    ) )
    {
      System.err.println ( "[DataTaskArgument]  Error:  "
			   + "Invalid Argument Direction ("
			   + theArgumentDirection + ").  Assuming default." );
      argumentDirection = DataTaskArgument . DEFAULT_DIRECTION;
    }
    else
    {
      argumentDirection = theArgumentDirection;
    }
  }


  public String getArgumentDirectionString ( )
  {
    return DataTaskArgument . ARGUMENT_TYPES [ getArgumentDirection() ];
  }

  public void   setArgumentDirectionString ( String theArgumentDirection )
  {
    int   newDirection = DataTaskArgument . getArgumentDirectionOfString (
							theArgumentDirection );

    if ( newDirection == DataTaskArgument . INVALID_DIRECTION )
    {
      System.err.println ( "[DataTaskArgument:setArgumentDirectionString]  "
			   + "Warning:  Invalid direction (\""
			   + theArgumentDirection
			   + "\".  Using default direction instead." );
      newDirection = DataTaskArgument . DEFAULT_DIRECTION;
    }

    setArgumentDirection ( newDirection );
  }


  public  String getArgumentName ( ) { return argumentName; }

  protected void setArgumentNameWithoutParsingChecks ( String theArgumentName )
				     { argumentName = theArgumentName; }

  public    void setArgumentName ( String theArgumentName )
    throws DetailedParseException
  {
	/* Special case blank/empty name */
    if ( DataComponent . isEmptyString ( theArgumentName ) )
    {
      setArgumentNameWithoutParsingChecks ( DataComponent.EMPTY_STRING );
      return;
    }

	/* General case -- parse the name */
    TDLParser . reinitParser ( theArgumentName );
    try
    {
      setArgumentNameWithoutParsingChecks ( 
			       TDLParser . getParser() . parseArgumentName() );
    }
    catch ( Throwable  theExceptionOrError )
    {
      didParseOfSubpartFail ( theExceptionOrError );
    }
  }


  public boolean hasEquals()
  {
    return hasIndex ( DataTaskArgument.ARGUMENT_EQUALS_INDEX );
  }


  public int countSubcomponentNewlinesAfterEquals()
  {
    return countSubcomponentNewlines (
		       getIndex ( DataTaskArgument.ARGUMENT_EQUALS_INDEX ) + 1,
		       getSubcomponentsCount(),
		       DataComponent.ENTIRE_OBJECT,
		       true /* Stop at first non-whitespace character */ );
  }



  public String getWarnString ( int      theObjectSubset,
				boolean  theIsGenerating )
  {
    String  baseString
              = super . getWarnString ( theObjectSubset, theIsGenerating )
                 + " or DataTaskArgument.TYPE_NAME_AND_EQUALS ("
                 + DataTaskArgument.TYPE_NAME_AND_EQUALS + ")";

    if ( theIsGenerating )
      return baseString
                 + " or DataTaskArgument.TYPE_AND_NAME ("
                 + DataTaskArgument.TYPE_AND_NAME + ")"
	         + " or DataTaskArgument.AFTER_EQUALS ("
		 + DataTaskArgument.AFTER_EQUALS + ")";
    else
      return baseString;

  }

  public boolean isValidObjectSubset ( int      theObjectSubset,
				       boolean  theIsGenerating )
  {
    if (   (    ( theIsGenerating == false )
	    &&  ( theObjectSubset == DataTaskArgument.TYPE_NAME_AND_EQUALS )
	    )
	|| (    ( theIsGenerating == true )
	    &&  (   ( theObjectSubset == DataTaskArgument.TYPE_NAME_AND_EQUALS)
	         || ( theObjectSubset == DataTaskArgument.TYPE_AND_NAME       )
		 || ( theObjectSubset == DataTaskArgument.AFTER_EQUALS        )
	         )
	    )
	)
      return true;
    else
      return super . isValidObjectSubset ( theObjectSubset, theIsGenerating );
  }


  public boolean isValid ( int theObjectSubsetToValidate )
  {
    	/* Print a warning if necessary... */
    warnIfInvalidObjectSubset ( theObjectSubsetToValidate, "isValid", false );

	/* Should have some way of checking to make sure a type has been set */
        /* Possibly re-parse this object??? */

    if ( DataComponent . isEmptyString ( getArgumentName() ) )
      return false;
    else
      return true;
  }


  protected
    DataComponent parseSpecificComponent ( TDLParser theTDLParser,
					   int       theObjectSubsetToParse )
    throws Exception
  {
    warnIfInvalidObjectSubset ( theObjectSubsetToParse,
				"parseSpecificComponent", false );

    if ( theObjectSubsetToParse == DataTaskArgument.TYPE_NAME_AND_EQUALS )
      return
	theTDLParser . parseTaskArgument_WithoutArgDir_AndTrailingEOF ( null );
    else
      return theTDLParser . parseTaskArgument_AndTrailingEOF ( null );
  }


	/* Actual work of reseting this DataObject's data. */
  public boolean setFrom ( DataComponent  theObjectToCopy,
			   int            theObjectSubsetToSet,
			   boolean        theCanDestroyObject  )
  {
    if ( /*NOT*/ !  ( theObjectToCopy instanceof DataTaskArgument ) )
    {
      System.err.println ( "[DataTaskArgument:setFrom]  Error:  "
			   + "Illegal object.  (\""
			   + theObjectToCopy . getClass() . getName()
			   + "\")" );
      return false;
    }

    DataTaskArgument  theArgumentToCopy = (DataTaskArgument) theObjectToCopy;

    isPersistentDeclaration = theArgumentToCopy.isPersistentDeclaration;
    argumentDirection       = theArgumentToCopy.argumentDirection;
    argumentName            = theArgumentToCopy.argumentName;


	/* Copy our leading subcomponents... */
    if ( theObjectSubsetToSet == DataTaskArgument.TYPE_NAME_AND_EQUALS )
    {
      int     i;
      int     ourDirectionIndex = getIndex (
				   DataTaskArgument.ARGUMENT_DIRECTION_INDEX );
      int     newDirectionIndex;

      tmpVector . removeAllElements();

      for ( i=0;  i < ourDirectionIndex;  i++ )
      {
	tmpVector . addElement ( getSubcomponent ( i ) );
      }

      if ( super . setFrom ( theObjectToCopy, theObjectSubsetToSet,
			     theCanDestroyObject )
	   == false )
      {
	return false;
      }

      newDirectionIndex
	= getIndex ( DataTaskArgument.ARGUMENT_DIRECTION_INDEX );

      for ( i=0;  i < newDirectionIndex;  i++ )
	removeSubcomponent ( 0 );

      for ( i=0;  i < ourDirectionIndex;  i++ )
      {
	privateAddSubcomponent ( tmpVector . elementAt ( i ), i );
      }

      setIndex ( DataTaskArgument.ARGUMENT_DIRECTION_INDEX,
		 ourDirectionIndex );

      setIndex ( DataTaskArgument.ARGUMENT_NAME_INDEX,
		 getIndex ( DataTaskArgument.ARGUMENT_NAME_INDEX )
		 - newDirectionIndex + ourDirectionIndex );

      setIndex ( DataTaskArgument.ARGUMENT_EQUALS_INDEX,
		 getIndex ( DataTaskArgument.ARGUMENT_EQUALS_INDEX )
		 - newDirectionIndex + ourDirectionIndex );

      return true;
    }
    else
    {
      return super . setFrom ( theObjectToCopy, theObjectSubsetToSet,
			       theCanDestroyObject );
    }
  }



  public void generate ( DataDestination  theOutputDestination,
			 int              theObjectSubsetToGenerate )
  {
      /* Print a warning if necessary... */
    warnIfInvalidObjectSubset ( theObjectSubsetToGenerate, "generate", true );

	/* If we are generating the argument-direction... */
    if (   (theObjectSubsetToGenerate != DataTaskArgument.TYPE_NAME_AND_EQUALS)
	&& (theObjectSubsetToGenerate != DataTaskArgument.TYPE_AND_NAME       )
	&& (theObjectSubsetToGenerate != DataTaskArgument.AFTER_EQUALS        )
	)
    {
	/* Initialize us to generate non-significant tokens... */
      initializeGenerateSubcomponentIndex();

	/* Write any pre-first-token non-significant tokens */
      generateSubcomponents ( DataComponent.FIRST_TOKEN_INDEX,
			      theOutputDestination,
			      theObjectSubsetToGenerate, false );

	/* Lets only do this if we are a PERSISTENT declaration */
      if ( getIsPersistentDeclaration() == true )
      {
	  /* Write any pre-arg-direction non-significant tokens */
	generateSubcomponents ( DataTaskArgument.PERSISTENT,
				theOutputDestination,
				theObjectSubsetToGenerate, false );

	  /* Write the argument-direction */
	theOutputDestination . write ( DataTaskArgument.PERSISTENT );
      }

	/* Lets only do this if we have an argument-direction!!! */
      if ( getArgumentDirection() != DataTaskArgument.NO_DIRECTION )
      {
	  /* Write any pre-arg-direction non-significant tokens */
	generateSubcomponents ( DataTaskArgument.ARGUMENT_DIRECTION_INDEX,
				theOutputDestination,
				theObjectSubsetToGenerate, false );

	  /* Write the argument-direction */
	theOutputDestination . write ( getArgumentDirectionString() );
      }
    }
    else /* Ie: No PERSISTENT keyword or argument-direction */
    {
	/* Initialize us to generate non-significant tokens... */
      initializeGenerateSubcomponentIndex (
				   DataTaskArgument.ARGUMENT_DIRECTION_INDEX );
    }

	/* Standard straightforward generation */
    if ( theObjectSubsetToGenerate != DataTaskArgument.AFTER_EQUALS )
    {
	/* Write any pre-arg-name (which includes the arg-type) tokens */
      generateSubcomponents ( DataTaskArgument.ARGUMENT_NAME_INDEX,
			      theOutputDestination,
			      theObjectSubsetToGenerate, false );

	/* write the argument-name */
      theOutputDestination . write ( getArgumentName() );


	/* If we are ending early...  Before the Equals... */
      if (   ( theObjectSubsetToGenerate == DataTaskArgument.TYPE_AND_NAME )
	  && ( hasEquals()                                                 ) )
      {
	generateSubcomponents ( DataTaskArgument.ARGUMENT_EQUALS_INDEX,
				theOutputDestination,
				theObjectSubsetToGenerate, false );
      }
      else
      {
	    /* Just write any remaining non-significant tokens */
	    /* Which includes any arg-equals tokens...         */
	generateAllRemainingSubcomponents ( theOutputDestination,
					    theObjectSubsetToGenerate, false );
      }
    }

    else /* Ie: theObjectSubsetToGenerate == DataTaskArgument.AFTER_EQUALS */
    {
	/* Generate all tokens after the arg-equals */
      initializeGenerateSubcomponentIndex (
		     getIndex ( DataTaskArgument.ARGUMENT_EQUALS_INDEX ) + 1 );
      generateAllRemainingSubcomponents ( theOutputDestination,
					  theObjectSubsetToGenerate, false );
    }
  }

}
