//*ALERT-DBA*: Does windows use a '/' or a '\' for the URL/HTML path separator?

import netscape.application.*;
import netscape.util.*;
import java.io.*;
import java.net.URL;
import java.net.MalformedURLException;


   /** Subclass this here to add/fix a little functionality */
public class ExtendedTextView extends TextView
{
	/* Constants */
		/* Note:  URL_PATH_SEPARATOR.length == 1 is assumed later on.*/
  public  final static String URL_PATH_SEPARATOR = "/";


	/* Target-Commands */
  public  final static String DELETE       = DefaultTextFilter.DELETE_COMMAND;
  public  final static String PREVIOUS_VIEW
				     = DefaultTextFilter.PREVIOUS_VIEW_COMMAND;
  public  final static String NEXT_VIEW   
				     = DefaultTextFilter.NEXT_VIEW_COMMAND;
	/* Note:  use delayedRedraw() to activate this target */
  public  final static String DELAYED_REDRAW = "DELAYED_REDRAW";
	/* Note:  Use delayedSelectRange() to activate this target */
  private final static String SELECT_RANGE = "SELECT_RANGE";
	/* Noted:  Hack used to check EventQueue Status... */
  private final static String NULL_COMMAND = "NULL_COMMAND";

  protected final static int MINIMUM_TEXT_VIEW_PADDING = 10;


	/* These next three instance variables form a truly hideous kludge.
	 * They serve to let ExtendedTextView perform delayed redrawing,
	 * which will occur after every regular redraw to avoid the
	 * TextView-turning-black bug.  Unfortunately, sometimes the
	 * redraws force new redraws, which can generate an extended queue
	 * of delayed-redraw-requests.  Or, alternatively, a delayed-redraw
	 * could trigger a regular-redraw, causing a cycle.  For this reason,
	 * a delayed-redraw count is kept, as well as an Event-Queue kludge
	 * to see if the current redraw was generated by a delayed-redraw...
	 * Yes, it's gross.  But it works...
	 */
  protected int     pendingDelayedRedraw;
  protected boolean doingDelayedRedraw;
  protected Event   kludgeEventLoopsNextEvent;

  protected int     alreadyDoingSizeBy;
  protected int     pendingSelectRangeCount;
  protected Range   aRange;
  protected int     cursor;
  protected boolean overridingCursor;

  public ExtendedTextView ( )
  {
    super();
    privateInitialize();
  }
  public ExtendedTextView ( Rect theRect )
  {
    super ( theRect );
    privateInitialize();
  }
  public ExtendedTextView ( int theX, int theY, int theWidth, int theHeight )
  {
    super ( theX, theY, theWidth, theHeight );
    privateInitialize();
  }

  private void privateInitialize()
  {
    alreadyDoingSizeBy        = 0;
    pendingSelectRangeCount   = 0;
    pendingDelayedRedraw      = 0;
    doingDelayedRedraw        = false;
    kludgeEventLoopsNextEvent = null;
    aRange = new Range();
    cursor = 0;
    overridingCursor = false;
    DefaultTextFilter.setDefaultTextFilter ( this );
  }

  public void drawView ( Graphics theGraphics )
  {
    super . drawView ( theGraphics );

	/* Hack to solve TextViews-turning-Black bug.
	 * Do a single delayed-redraw after every normal drawView...
	 */

    if ( doingDelayedRedraw )
    {
      doingDelayedRedraw = false;
      if ( pendingDelayedRedraw > 0 )
	pendingDelayedRedraw --;
    }

	/* If there are no delayed-redraws already pending... */
    else if ( pendingDelayedRedraw == 0 )
    {
	/* If the current event-loop event was NOT a delayed-redraw event */
      if (    kludgeEventLoopsNextEvent
	   != Application.application() . eventLoop() . peekNextEvent() )
      {
        delayedRedraw();
      }
    }
  }


  public void kludgeToForceLineLayout ( )
  {
	/* Massive Hack to get TextView to layout correctly... */
    replaceRangeWithString ( appendString ( " " ), "" );
	/* This next line is *pure* black magic of darkest night */
	/* selectedRange() should have been reset... above */
    selectRange ( selectedRange() );
  }

  public void delete()
  {
    if ( isEditable() )
    {
      Range range = selectedRange();
      if ( range != Range.nullRange() )
      {
	if ( range.length <= 0 )
	{
	  range.length = 1;
	}
	replaceRangeWithString(range, "");
	range.length = 0;
	selectRange ( range );
	sizeToMinSize();
      }
    }
  }

  public void cut()
  {
    if ( selectedRange() . length > 0 )
      super.cut();
  }

  public void copy()
  {
    if ( selectedRange() . length > 0 )
      super.copy();
  }

  public void paste()
  {
    super . paste();
    kludgeToForceLineLayout();
  }

  public void delayedRedraw()
  {
    Application . application() . performCommandLater ( this,
					       ExtendedTextView.DELAYED_REDRAW,
					       null, false );
    pendingDelayedRedraw ++;
  }

  public void delayedSelectRange ( Range theRange )
  {
    Application . application() . performCommandLater ( this,
					         ExtendedTextView.SELECT_RANGE,
						 theRange, false );
    pendingSelectRangeCount++;
  }

  public int getPendingSelectRangeCount()  { return pendingSelectRangeCount; }


  public boolean canPerformCommand ( String theCommand )
  {
    if (   ( super . canPerformCommand ( theCommand ) )
	|| (    theCommand . equals ( ExtendedTextView.DELETE )
	     && isEditable() )
	|| ( theCommand . equals ( ExtendedTextView.SELECT_RANGE   ) )
	|| ( theCommand . equals ( ExtendedTextView.PREVIOUS_VIEW  ) )
	|| ( theCommand . equals ( ExtendedTextView.NEXT_VIEW      ) )
	|| ( theCommand . equals ( ExtendedTextView.DELAYED_REDRAW ) )
	|| ( theCommand . equals ( ExtendedTextView.NULL_COMMAND   ) ) )
    {
      return true;
    }
    else
    {
      return false;
    }
  }

  public void performCommand ( String theCommand, Object theData )
  {
    if ( theCommand . equals ( ExtendedTextView.DELETE ) )
    {
      delete();
    }

    else if ( theCommand . equals ( ExtendedTextView.SELECT_RANGE ) )
    {
      setFocusedView();

      if ( theData instanceof Range )
	selectRange ( (Range) theData );
      else
      {
	aRange.index  = 0;
	aRange.length = length();
	selectRange ( aRange );
      }

      if ( pendingSelectRangeCount > 0 )
	pendingSelectRangeCount --;
    }

    else if ( theCommand . equals ( ExtendedTextView.PREVIOUS_VIEW ) )
    {
      rootView() . setFocusedView   ( null );
      rootView() . selectViewBefore ( this );
    }

    else if ( theCommand . equals ( ExtendedTextView.NEXT_VIEW ) )
    {
      rootView() . setFocusedView  ( null );
      rootView() . selectViewAfter ( this );
    }

    else if ( theCommand . equals ( ExtendedTextView.DELAYED_REDRAW ) )
    {
	  /* Can't check the event-loop for the current event that is being
	   * processed.  So, we have to check the event-loop's next event.
	   * Which is difficult if the event-loop is empty...
	   */
      if ( Application.application() . eventLoop() . peekNextEvent() == null )
      {
	  /* So put an event into the event-loop... */
	Application.application() . performCommandLater ( this,
					         ExtendedTextView.NULL_COMMAND,
					         null, false );
      }
      kludgeEventLoopsNextEvent
	= Application.application() . eventLoop() . peekNextEvent();
      doingDelayedRedraw = true;
      setDirty ( true );
    }

    else if ( theCommand . equals ( ExtendedTextView.NULL_COMMAND ) )
    {
      ; /* NULL/EMPTY STATEMENT.  Ie: NO-OP */
    }

    else
    {
      super . performCommand ( theCommand, theData );
    }
  }


	/* This next two bits -- doReformatAll and sizeBy are, quite possibly
	/* my grossest hacks of all time.  Lets hope they keep working. */
  public void doReformatAll()
  {
    alreadyDoingSizeBy++;
    didSizeBy ( 0, 0 );
    alreadyDoingSizeBy--;
  }
  public void sizeBy ( int deltaWidth, int deltaHeight )
  {
    alreadyDoingSizeBy++;
    super . sizeBy ( deltaWidth, deltaHeight );
    alreadyDoingSizeBy--;
    if ( alreadyDoingSizeBy == 0 )
      doReformatAll();
  }

  public void shrinkText()
  {
    sizeToMinSize();

    Range  totalRange = new Range ( 0, length() );
    Vector rects      = rectsForRange ( totalRange );
    Rect   tmpRect;
    int    i, rectWidth, width=1, height=1;

    for ( i=0;  i < rects.count();  i++ )
    {
      tmpRect = (Rect) rects . elementAt ( i );

      if ( tmpRect.width < 0 )
	rectWidth = bounds().width - tmpRect.width;
      else
	rectWidth = tmpRect.width;

      if ( i == 0 )
      {
	width  = tmpRect.width;
	height = tmpRect.height;
      }
      else
      {
	width   = Math.max ( width, tmpRect.width );
	height += tmpRect.height;
      }
    }

    width += ExtendedTextView.MINIMUM_TEXT_VIEW_PADDING;

    sizeTo ( width, height );
  }


  public void overrideCursor ( int theNewCursor )
  {
    overridingCursor = true;
    cursor = theNewCursor;
  }

  public void resumeCursor ( )
  {
    overridingCursor = false;
    cursor = 0;
  }

  public int cursorForPoint ( int theX, int theY )
  {
    if ( overridingCursor )
      return cursor;
    else
      return super.cursorForPoint ( theX, theY );
  }

  public InputStream generateHTMLInputStream ( File theFile )
    throws FileNotFoundException
  {
    StringBuffer  stringBuffer;
    String        directoryList[];
    int           i;

	/* If this is a file, or this file does not exist */
    if ( theFile . isDirectory() == false )
    {
      return new BufferedInputStream ( new FileInputStream ( theFile ) );
    }

    stringBuffer = new StringBuffer();
    stringBuffer . append ( "<HTML>\n<HEAD>\n<TITLE>Directory:  " );
    stringBuffer . append ( theFile.getAbsolutePath() );
    stringBuffer . append ( "</TITLE>\n</HEAD>\n<BODY>\n\n" );

    stringBuffer . append ( "<H1 ALIGN=CENTER>Directory:  " );
    stringBuffer . append ( theFile.getAbsolutePath() );
    stringBuffer . append ( "</H1><HR><BR>\n\n<PRE>\n" );


    directoryList = theFile . list();
    Sort.sortStrings ( directoryList, 0, directoryList.length, true, false );
    for ( i=0;  i < directoryList.length;  i++ )
    {
      if ( ( new File ( theFile, directoryList[i] ) ) . isDirectory() )
	stringBuffer . append ( "(<B>DIR</B>)  " );
      else
	stringBuffer . append ( "(FILE) " );
      stringBuffer . append ( "<A HREF=\"" );
      stringBuffer . append ( directoryList[i] );
      stringBuffer . append ( "\">" );
      stringBuffer . append ( directoryList[i] );
      stringBuffer . append ( "</A>\n" );
    }
    stringBuffer . append ( "</PRE>\n\n</BODY>\n</HTML>\n" );

    return new StringBufferInputStream ( stringBuffer.toString() );
  }



  public void importHTMLFromURL ( String theURLString )
    throws MalformedURLException, IOException, HTMLParsingException

  {
    String       filepathSeparator;
    File         urlFile;
    InputStream  urlInputStream = null;
    URL          aURL = new URL ( theURLString );

	/* Start it with "file:" */
    if ( theURLString . startsWith ( "/" ) )
      theURLString = "file:" + theURLString;

	/* Hack directories to make them look nice... */
    if ( theURLString . startsWith ( "file:" ) )
    {
      filepathSeparator
	= System.getProperty ( "file.separator",
			       ExtendedTextView.URL_PATH_SEPARATOR );

      if ( ExtendedTextView.URL_PATH_SEPARATOR . equals ( filepathSeparator )
	   == false )
      {
	if ( filepathSeparator . length() != 1 )
	{
	  urlFile = null;
	}
	else
	{
	  urlFile = new File ( theURLString . substring ( 5 ) . replace (
					  ExtendedTextView.URL_PATH_SEPARATOR
					    . charAt ( 0 ),
					  filepathSeparator . charAt ( 0 ) ) );
	}
      }
      else
      {
	urlFile = new File ( theURLString . substring ( 5 ) );
      }


      if (   ( urlFile != null    )
	  && ( urlFile . exists() ) )
      {
	try
	{
	  urlInputStream = generateHTMLInputStream ( urlFile );
	}
	catch ( FileNotFoundException theException )
	{
	  urlInputStream = null;
	}
      }
    }

    if ( urlInputStream == null )
      urlInputStream = aURL. openStream();

    aRange . index  = 0;
    aRange . length = length();
    importHTMLInRange( urlInputStream , aRange , aURL );

    aRange.index = 0;
    aRange.length = 0;
    selectRange          ( aRange );
    scrollRangeToVisible ( aRange );
  }


  public void importHTMLFromString ( String theHTMLString )
  {
    if ( length() == 0 )
      setString ( " " );

    aRange . index  = 0;
    aRange . length = length();

    try
    {
      importHTMLInRange ( new StringBufferInputStream ( theHTMLString ),
			  aRange , null );
    }
    catch ( HTMLParsingException  theException )
    {
      System.err.println ( "[ExtendedTextView:importHTMLFromString]  Error:  "
		   + "HTML Parsing Error at line " + theException.lineNumber()
		   + ":    " + theException.toString() );
      return;
    }
    catch ( IOException  theException )
    {
      System.err.println ( "[ExtendedTextView:importHTMLFromString]  Error:  "
			  + "IOException Error:  " + theException.toString() );
      return;
    }

    aRange.index = 0;
    aRange.length = 0;
    selectRange          ( aRange );
    scrollRangeToVisible ( aRange );
  }

}
