
import netscape.application.*;
import netscape.util.*;

public class PaletteView extends SimpleLabel implements DragSource,
							GraphComponentSupplier
{
	/* Class Variables/Constants */
  public final static int MOUSE_X_DRAG_SESSION_FUDGE_FACTOR
			    = GraphComponent.MOUSE_X_DRAG_SESSION_FUDGE_FACTOR;

	/* Class variable, used to simplify bitmap construction */
  protected static Bitmap invisiblePixelBitmap = null;

	/* Class method, used to simplify bitmap construction */
  public static void makeBitmapInvisible ( Bitmap   theBitmap,
					   Graphics theGraphics )
  {
    final int  stepSize = 20;
    int        i,w,h;
    Graphics   bitmapGraphics;

	/* Make sure we have an invisible-bits image to utilize */
    if ( PaletteView.invisiblePixelBitmap == null )
    {
      int   pixels[] = new int [ stepSize * stepSize ];
      for ( i=0;  i < (stepSize * stepSize);  i++ )
	pixels [ i ] = 0x00ff0000; /* If they show up red, ALPHA is broken. */
      PaletteView.invisiblePixelBitmap = new Bitmap ( pixels,
						      stepSize, stepSize );
    }

	/* Make certain that we have Graphics */
    if ( theGraphics == null )
      bitmapGraphics = theBitmap . createGraphics();
    else
      bitmapGraphics = theGraphics;

	/* This would help... */
    theBitmap . setTransparent ( true );

	/* Draw the invisible-bits bitmap all over the new bitmap */
    for (   w=0;  (w * stepSize) < theBitmap.width();   w++ )
    {
      for ( h=0;  (h * stepSize) < theBitmap.height();  h++ )
      {
	bitmapGraphics . drawBitmapAt ( PaletteView.invisiblePixelBitmap,
					(w * stepSize), (h * stepSize) );
      }
    }

	/* If we created our graphics, delete it */
    if ( theGraphics != bitmapGraphics )
      bitmapGraphics . dispose();

    return;
  }



	/* Instance Variables */
  protected  Bitmap         image;
  protected  Color          color;
  protected  DataComponent  dataComponent;
  protected  boolean        hasInitialized;/*Hack to deal with initialization*/

  public PaletteView ( String         theName,
		       DataComponent  theDataComponent,
		       Color          theColor )
  {
    super();
    color = theColor;
    setInternalDataComponent ( theDataComponent );
    hasInitialized = true;
    setText ( theName );
  }


  public void setText ( String theText )
  {
    super . setText ( theText );

        /* Abort if we haven't finished initializing yet... */
    if ( hasInitialized == false )
      return;

    Graphics  imageGraphics;
    Color     originalBackgroundColor = getBackgroundColor();
    int       i,
              textWidth, textHeight,
              targetHeight, targetWidth,
              borderSize = 2,
              lineSize   = GraphConnection.DEFAULT_HORIZONTAL_LENGTH;

	/* Hack bounds() to allow room for a border & left line */
    textWidth    = bounds() . width;
    textHeight   = bounds() . height;
    targetWidth  = textWidth  + ( 2 * borderSize ) + lineSize;
    targetHeight = textHeight + ( 2 * borderSize );

	/* Create the image & graphics for drawing to the image */
    image         = new Bitmap ( targetWidth, targetHeight );
    imageGraphics = image . createGraphics();

	/* Make the image invisible */
/* Disable this until invisible bitmaps work right in dragviews */
//    PaletteView.makeBitmapInvisible ( image , imageGraphics );


	/* Use superclass to simplify writing the text... */
    imageGraphics . translate ( borderSize + lineSize, borderSize );
    setBackgroundColor ( null );
    super . drawView ( imageGraphics );
    setBackgroundColor ( originalBackgroundColor );
    imageGraphics . translate ( - borderSize - lineSize, - borderSize );



	/* Color these next lines appropriately */
    imageGraphics . setColor ( getColor() );

	/* Draw the lines. */
    imageGraphics . drawLine ( 0,        targetHeight / 2,
			       lineSize, targetHeight / 2 );
    for ( i=0;  i < borderSize;  i++ )
    {
      imageGraphics . drawRect ( i + lineSize,
				 i,
				 targetWidth  - (2 * i) - lineSize,
				 targetHeight - (2 * i) );
    }

    imageGraphics . dispose();

    sizeTo     ( targetWidth, targetHeight );
    setMinSize ( targetWidth, targetHeight );
  }


  public Bitmap getImage       ( ) { return image;       }

  public Color getColor() { return color; }
  public void  setColor ( Color theColor )
  {
    color = theColor;
    setText ( getText() );
  }

  public DataComponent getInternalDataComponent() { return dataComponent; }
  public void          setInternalDataComponent(DataComponent theDataComponent)
					  { dataComponent = theDataComponent; }



  public void drawView ( Graphics theGraphics )
  {
    theGraphics . setColor ( getBackgroundColor() );
    theGraphics . fillRect ( 0, 0,
			     bounds() . width,
			     bounds() . height );

    if ( getImage() != null )
      theGraphics . drawBitmapAt ( getImage(), 0, 0 );
  }


  public boolean mouseDown ( MouseEvent theEvent )
  {
    new DragSession ( this, getImage(),
		      theEvent.x
			  + PaletteView.MOUSE_X_DRAG_SESSION_FUDGE_FACTOR,
		      theEvent.y
			  - ( getImage() . height()  /  2 ),
		      theEvent.x, theEvent.y,
		      "PaletteViewSource",
		      this );
    return false;
  }

	/* Drag source interface */
  public void dragWasAccepted ( DragSession theDragSession )
  {
  }
  public boolean dragWasRejected ( DragSession theDragSession )
  {
    return true;
  }
  public View sourceView ( DragSession theDragSession )
  {
    return this;
  }


	/* GraphComponentSupplier interface */
  public boolean supportsGraphType ( int theGraphType )
  {
    if ( getInternalDataComponent() != null )
      return getInternalDataComponent() . supportsGraphType ( theGraphType );
    else
    {
      System.err.println ( "[PaletteView:supportsGraphType]  Warning:  "
			   + "getInternalDataComponent() returns null." );
      return false;
    }
  }

  public GraphModule createGraphModule ( int theGraphType )
  {
    if ( getInternalDataComponent() != null )
    {
      return getInternalDataComponent() . createGraphModule ( theGraphType );
    }
    else
    {
      System.err.println ( "[PaletteView:createGraphModule]  Warning:  "
			   + "getInternalDataComponent() returns null." );
      return null;
    }
  }

	/* Actual work of creating a new DataComponent Object */
  public DataComponent getDataComponent  (  )
  {
    DataComponent   dataComponent;

    if ( getInternalDataComponent() != null )
    {
      return getInternalDataComponent() . newComponent();      
    }
    else
    {
      System.err.println ( "[PaletteView:getDataComponent]  Warning:  "
			   + "getInternalDataComponent() returns null." );
    }
    return null;
  }

}
