
import netscape.application.*;
import netscape.util.*;
 

/** The ScrollText class is designed to wrap a TextView inside a
  * ScrollView such that the TextView will put each paragraph on it's
  * own horizontal line.  (Extending it's size as needed to fit the line. )
  *
  * Note:  ScrollText uses an enhanced TextView called ExtendedTextView.
  */
public class ScrollText extends ScrollGroup implements TextViewOwner
{
  protected boolean           showEntireText, isAutoResizing;
  protected View              mouseClient;
  protected ExtendedTextView  ourTextView;
  protected Size              ourMinSize, actualMinSize;


  public ScrollText ( )  { this ( 0, 0, 1, 1 ); }

  public ScrollText ( Rect theRect ) 
	    { this ( theRect.x, theRect.y, theRect.width, theRect.height ); }

  public ScrollText ( int theX, int theY, int theWidth, int theHeight )
  {
    super ( theX, theY, theWidth, theHeight );

    showEntireText = false;
    isAutoResizing = true;
    mouseClient    = null;

    ourTextView = new ExtendedTextView ( theX, theY, theWidth, theHeight );
    ourTextView . setOwner ( this );

    setContentView ( ourTextView );

    setBorder ( null );
    setHorizScrollBarDisplay ( ScrollGroup . AS_NEEDED_DISPLAY );
    setVertScrollBarDisplay  ( ScrollGroup . AS_NEEDED_DISPLAY );
    setBackgroundColor ( getTextView() . backgroundColor() );

	/* Define our minimum-size. */
    ourMinSize    = new Size ( 0, 0 );
    actualMinSize = super.minSize();
	/* IFC Bug fix/kludge for minimum-size */
    actualMinSize . width  += horizScrollBar() . minKnobLength();
    actualMinSize . height += vertScrollBar()  . minKnobLength();
  }

  public boolean getShowEntireText ( ) { return showEntireText; }
  public void    setShowEntireText ( boolean theShowEntireText )
			               { showEntireText = theShowEntireText; }

  public boolean getIsAutoResizing ( ) { return isAutoResizing; }
  public void    setIsAutoResizing ( boolean theIsAutoResizing )
				    { isAutoResizing = theIsAutoResizing; }

  public ExtendedTextView getTextView ( ) { return ourTextView; }


  public void setTextViewSize ( int theWidth, int theHeight )
  {
    getTextView() . sizeBy ( theWidth  - getTextView().bounds().width,
			     theHeight - getTextView().bounds().height );
    getTextView() . sizeToMinSize();
    getTextView() . setDirty ( true );
  }


  public String getText ( ) { return getTextView() . string(); }

  public void setText ( String theText )
  {
    getTextView() . setString ( "" );
    appendText ( theText );
  }

  public void appendText ( String theText )
  {
    int    start=0, end=0;

	/* TextView's do not handle "\n"'s in strings gracefully... */
    while ( end < theText . length() )
    {
      end = theText . indexOf ( "\n", start );
      if ( end == -1 )
	end = theText . length();

      getTextView() . appendString ( theText.substring ( start, end ) );

      if (   ( end < theText . length() )
	  && ( theText . charAt ( end ) == '\n' ) )
      {
	getTextView() . appendString ( "\n" );
	end ++;
      }

      start = end;
    }

	  /* Massive Hack to get TextView to layout correctly... */
    getTextView() . kludgeToForceLineLayout();
	/* Better reset our size, now... */
    resize();
  }

  public int lengthOfText ( ) { return getTextView() . length(); }

  public Font getFont ( ) { return getTextView() . font(); }

  public void setFont ( Font theFont ) { getTextView() . setFont ( theFont ); }


  public Size getActualMinScrollSize ( ) { return actualMinSize; }


  public Size minSize ( )
  {
    int   horizontalSize = border().widthMargin(),
          verticalSize   = border().heightMargin(),
          HorizontalScrollbarWidth  = (   horizScrollBar().minSize().width
					+ horizScrollBar().minKnobLength() ),
          HorizontalScrollbarHeight = horizScrollBar().minSize().height,
          VerticalScrollbarHeight   = (   vertScrollBar().minSize().height
				        + vertScrollBar().minKnobLength() ),
          VerticalScrollbarWidth    = vertScrollBar().minSize().width,
          minWidth  = HorizontalScrollbarWidth + VerticalScrollbarWidth,
          minHeight = VerticalScrollbarHeight  + HorizontalScrollbarHeight;

	/* No Scrollbars needed. */
    if      (   ( getTextView().bounds().width  <= minWidth  )
	     && ( getTextView().bounds().height <= minHeight ) )

    {
      horizontalSize += getTextView().bounds().width;
      verticalSize   += getTextView().bounds().height;
    }

	/* Horizontal Scrollbar needed */
    else if (   ( getTextView().bounds().width  >  minWidth                )
	     && ( getTextView().bounds().height <= VerticalScrollbarHeight ) )
    {
      horizontalSize +=     HorizontalScrollbarWidth;
      verticalSize   += (   HorizontalScrollbarHeight
			  + getTextView().bounds().height );
    }

	/* Vertical Scrollbar needed */
    else if (   ( getTextView().bounds().width  <= HorizontalScrollbarWidth )
	     && ( getTextView().bounds().height >  minHeight                ) )
    {
      horizontalSize += (   VerticalScrollbarWidth
			  + getTextView().bounds().width );
      verticalSize   += VerticalScrollbarHeight;
    }

	/* Both Scrollbars needed */
    else 
    {
      horizontalSize = getActualMinScrollSize() . width;
      verticalSize   = getActualMinScrollSize() . height;
    }

    ourMinSize . width  = horizontalSize;
    ourMinSize . height = verticalSize;

    return ourMinSize;
  }


  public void sizeToTextSize ( )
  {
    sizeTo ( getTextView() . bounds . width,
	     getTextView() . bounds . height );
  }

  public void setToMinScrollSize ( )
  {
    int     width  = bounds() . width,
            height = bounds() . height;
    boolean didResize = false;

    if ( getActualMinScrollSize() . width > width )
    {
      width = getActualMinScrollSize() . width;
      didResize = true;
    }

    if ( getActualMinScrollSize() . height > height )
    {
      height = getActualMinScrollSize() . height;
      didResize = true;
    }

    if ( didResize )
      sizeTo ( width, height );
  }

  public void resize ( )
  {
    getTextView().shrinkText();

    if ( getIsAutoResizing() == false )
      return;

    Size  size = minSize();

    if ( getShowEntireText() )
    {
      size . width  =  getTextView() . bounds() . width;
      size . height =  getTextView() . bounds() . height;
    }

    if (   ( bounds().width  != size.width  )
	|| ( bounds().height != size.height ) )
    {
      sizeTo ( size.width, size.height );
    }
  }



	/* Pass mouse events that we don't want upward to our parents */
  public boolean mouseDown ( MouseEvent theEvent )
  {
    if ( super . mouseDown ( theEvent ) == true )
      return true;

	/* Do some dark-magic here to pass this mouseDown event up to */
	/* our parent, if they want it... */
    else if (   ( superview() != null )
	     && ( superview()
		    . mouseDown( convertEventToView ( superview(), theEvent ) )
		 ) )
    {
      mouseClient = superview();
      return true;
    }

    return false;
  }

  public void mouseDragged ( MouseEvent theEvent )
  {
    if ( mouseClient != null )
      mouseClient
	. mouseDragged ( convertEventToView ( mouseClient, theEvent ) );
  }

  public void mouseUp ( MouseEvent theEvent )
  {
    if ( mouseClient != null )
    {
      mouseClient . mouseUp ( convertEventToView ( mouseClient, theEvent ) );
      mouseClient = null;
    }
  }


	/* TextViewOwner Interface */
  public  void attributesDidChange(TextView t, Range r) {}
  public  void attributesWillChange(TextView t, Range r) {}
  public  void linkWasSelected(TextView t, Range r, String s) {}
  public  void textEditingDidBegin(TextView t) {}
  public  void textEditingDidEnd(TextView t) {}
  public  void selectionDidChange ( TextView theTextView ) {}

  public  void textWillChange ( TextView theTextView, Range theRange )
  {
	/* disable drawing to avoid "flash" */
    disableDrawing();
    theTextView.disableDrawing();
  }


  public  void textDidChange ( TextView theTextView, Range theRange )
  {
	/* Idiocy check */
    if ( theTextView != getTextView() )
    {
      System.err.println ( "[ScrollText:textDidChange]  Warning:  "
			   + "Not my TextView...  Aborting." );
	/* Reenable drawing... */
      reenableDrawing();
      theTextView.reenableDrawing();
	/* And return... */
      return;
    }

	/* If we have only made the text smaller, don't do anything... */
    if ( theRange . length() <= 0 )
    {
	/* Reenable drawing... */
      reenableDrawing();
      theTextView.reenableDrawing();
	/* And return... */
      return;
    }

	/* Otherwise, we have "grown" or otherwise "altered" the text. */
	/* So lets figure out our new "maximum" text width. */

    Range        paragraphRange,
	         runRange;
    String       textString;
    FontMetrics  textFontMetrics = null;
    int          rangeIndex      = theRange . index(),
	         paragraphIndex,
                 paragraphWidth,
                 maxWidthInRange = 0;

	/* For every paragraph contain (fully or partially) in theRange */
    while ( theRange . contains ( rangeIndex ) )
    {
      paragraphRange = theTextView . paragraphForIndex ( rangeIndex );
      
	/* If this is a zero-length paragraph, ignore it. */
      if ( paragraphRange.length() <= 0 )
      {
	rangeIndex++;
	continue;
      }

      paragraphWidth = 0;
      paragraphIndex = paragraphRange.index();

	/* For every RUN in this paragraph */
      while ( paragraphRange . contains ( paragraphIndex ) )
      {
	runRange = theTextView . runForIndex ( paragraphIndex );
	
	 /*If this is a zero-length run, something is wrong - try to recover.*/
	if ( runRange.length() <= 0 )
	{
	  System.err.println ( "[ScrollText:textDidChange]  Warning:  "
			       + "Zero-length run found.  Recovering..." );
	  paragraphIndex++;
	  continue;
	}

	if ( runRange.lastIndex() > paragraphRange.lastIndex() )
	{
	  System.err.println ( "[ScrollText:textDidChange]  Warning:  "
			       + "Run exceeds paragraph.  Using Run length." );
	}

		/* Add the width of this RUN to the paragraph's total width */
	textString = theTextView . stringForRange ( runRange );
	textFontMetrics
	  = ( (Font) ( theTextView . attributesAtIndex ( runRange.index() )
			           . get ( TextView.FONT_KEY ) )
	     ) . fontMetrics();
	paragraphWidth += textFontMetrics . stringWidth( textString );

		/* Go on to the next RUN */
	if ( runRange . contains ( paragraphIndex ) )
	  paragraphIndex = runRange . lastIndex()  + 1;
	else
	{
	  System.err.println ( "[ScrollText:textDidChange]  Warning:  "
			       + "RUN does not contain paragraphIndex." );
	  paragraphIndex ++;
	}

      } /* while ( paragraphRange . contains ( paragraphIndex ) ) */

	/* Is this our new widest paragraph? */
      maxWidthInRange = Math.max ( maxWidthInRange, paragraphWidth );


	/* Go on to the next paragraph.
	 *
	 * Note: It is possible that the character at rangeIndex is a '\n'
	 * terminating a paragraph, in which case the paragraphRange won't
	 * contain rangeIndex.  Lets not create an endless loop in this case.
	 */
      if ( paragraphRange . contains ( rangeIndex ) )
	rangeIndex = paragraphRange . lastIndex()  + 1;
      else
	rangeIndex ++;

    } /* while ( theRange . contains ( rangeIndex ) ) */



	/* Allow a the little extra padding room */
    if ( textFontMetrics != null )
      maxWidthInRange += textFontMetrics . stringWidth ( "AAA" );
    else
      maxWidthInRange += ExtendedTextView.MINIMUM_TEXT_VIEW_PADDING;

	/* And rescale if necessary */
    if ( maxWidthInRange > theTextView . bounds() . width )
    {
      setTextViewSize ( maxWidthInRange, theTextView . bounds() . height );
      theTextView . scrollRangeToVisible ( theRange );
    }

	/* And resize ourself if we need to. */
    resize();
  
	/* And reenable drawing so we can see what we have done. */
    reenableDrawing();
    theTextView.reenableDrawing();
	/* And return... */
    return;
  }

}
