
import netscape.application.*;
import netscape.util.*;
import netscape.constructor.*;
import java.io.File;
import java.net.MalformedURLException;
import java.io.IOException;
import java.io.InputStream;


public class TaskBrowserWindowHandler extends WindowHandlerAbstractBaseClass
			           implements ExtendedTarget, Target,
					      TextFieldOwner, TextViewOwner
{
	/* Constant (Special) Filenames */
  public final static String TASK_BROWSER_PLAN_NAME= "TaskBrowserWindow.planb";

	/* TASK_BROWSER_PLAN Objects */
  public final static String BACKWARD_BUTTON_NAME  = "BackButton";
  public final static String FORWARD_BUTTON_NAME   = "ForwardButton";
  public final static String RELOAD_BUTTON_NAME    = "ReloadButton";
  public final static String INTERNAL_EXTERNAL_POPUP_NAME
						   = "InternalExternalPopup";
  public final static String URL_TEXTFIELD_NAME    = "URLTextField";
  public final static String HTML_SCROLLGROUP_NAME = "HTMLScrollGroup";
  public final static String SEARCH_BUTTON_NAME    = "SearchButton";
  public final static String SEARCH_TEXTFIELD_NAME = "SearchTextField";
  public final static String SEARCH_DIRECTION_POPUP_NAME
						   = "SearchDirectionPopup";
  public final static String CASE_SENSITIVE_NAME   = "CaseSensitiveButton";
  public final static String SEARCH_DIRECTORY_NAME = "SearchDirectoryButton";


	/* Commands */
  public final static String FORWARD_COMMAND       = "ForwardHTML";
  public final static String BACKWARD_COMMAND      = "BackwardHTML";
  public final static String SEARCH_COMMAND        = "SearchHTML";
  public final static String RELOAD_COMMAND        = "ReloadHTML";
  public final static String REALLY_RELOAD_COMMAND = "ReallyReloadHTML";
  public final static String SEARCH_DIRECTORY_COMMAND
						   = "SearchDirectory";

	/* Constants */
  public final static int    INTERNAL_WINDOW_POPUP_INDEX = 0;
  public final static int    EXTERNAL_WINDOW_POPUP_INDEX = 1;
  public final static int    SEARCH_FORWARD_POPUP_INDEX  = 0;
  public final static int    SEARCH_BACKWARD_POPUP_INDEX = 1;

  public final static String TITLE           = "Task Browser";
  public final static int    MAX_HISTORY     = 50;
  public final static String PATH_SEPARATOR
				 = ExtendedTextView.URL_PATH_SEPARATOR;


	/* Class variables */
  protected static Plan      taskBrowserWindowPlan = null;

	/* Class Methods */
  protected static synchronized Plan getTaskBrowserWindowPlan()
  {
    if ( TaskBrowserWindowHandler.taskBrowserWindowPlan == null )
    {
      TaskBrowserWindowHandler.taskBrowserWindowPlan
	= WindowHandlerAbstractBaseClass.getPlan (
		TaskBrowserWindowHandler.TASK_BROWSER_PLAN_NAME,
		"[TaskBrowserWindowHandler:getTaskBrowserWindowPlan]  "
		+ "Error:  Failure - Unable to access TASK_BROWSER_PLAN_NAME");
    }
    return TaskBrowserWindowHandler.taskBrowserWindowPlan;
  }




	/* Instance Variables */
  protected Button            backwardButton,
		              forwardButton,
		              reloadButton,
		              searchButton,
			      caseSensitiveButton,
			      searchDirectoryButton;
  protected Popup             internalExternalPopup,
			      searchDirectionPopup;
  protected TextField         urlTextfield,
			      searchTextfield;
  protected ScrollGroup       htmlScrollgroup;
  protected ExtendedTextView  htmlTextview;

  protected Vector            urlVector;
  protected int               currentUrlIndex;
  protected Range             searchRange;

  protected SearchDirectoryWindowHandler  searchDirectoryWindowHandler;


	/* Constructors */
  public TaskBrowserWindowHandler ( VDT  theVDT )
		      { createAndInitialize ( false, theVDT  ); }

  public TaskBrowserWindowHandler ( boolean  theUseExternalWindow,
				    VDT      theVDT               )
		      { createAndInitialize ( theUseExternalWindow, theVDT ); }

	/* Constructors can't be synchronized. */
	/* And this needs to be synchronized since we use the static plans */
  private synchronized
     boolean createAndInitialize ( boolean theUseExternalWindow, VDT theVDT )
  {
	/* Initialize variables */
    urlVector       = new Vector();
    currentUrlIndex = 0;
    searchRange     = new Range();
    searchDirectoryWindowHandler = null;

	/* Let the super-class do all the work... */
    if ( super . createAndInitialize ( 
		     TaskBrowserWindowHandler.getTaskBrowserWindowPlan(),
		     theUseExternalWindow,
		     TaskBrowserWindowHandler.TITLE,
		     "TaskBrowserWindowHandler",
		     theVDT )
	 == false )
    {
      return false;
    }

	/* Lets find our instance variables */
    backwardButton
      = (Button)
         TaskBrowserWindowHandler.getTaskBrowserWindowPlan().componentNamed (
			     TaskBrowserWindowHandler . BACKWARD_BUTTON_NAME );
    forwardButton
      = (Button)
         TaskBrowserWindowHandler.getTaskBrowserWindowPlan().componentNamed (
			      TaskBrowserWindowHandler . FORWARD_BUTTON_NAME );
    reloadButton
      = (Button)
         TaskBrowserWindowHandler.getTaskBrowserWindowPlan().componentNamed (
			       TaskBrowserWindowHandler . RELOAD_BUTTON_NAME );
    searchButton
      = (Button)
         TaskBrowserWindowHandler.getTaskBrowserWindowPlan().componentNamed (
			       TaskBrowserWindowHandler . SEARCH_BUTTON_NAME );
    caseSensitiveButton
      = (Button)
         TaskBrowserWindowHandler.getTaskBrowserWindowPlan().componentNamed (
			      TaskBrowserWindowHandler . CASE_SENSITIVE_NAME );
    searchDirectoryButton
      = (Button)
         TaskBrowserWindowHandler.getTaskBrowserWindowPlan().componentNamed (
			    TaskBrowserWindowHandler . SEARCH_DIRECTORY_NAME );
    internalExternalPopup
      = (Popup)
         TaskBrowserWindowHandler.getTaskBrowserWindowPlan().componentNamed (
		     TaskBrowserWindowHandler . INTERNAL_EXTERNAL_POPUP_NAME );
    searchDirectionPopup
      = (Popup)
         TaskBrowserWindowHandler.getTaskBrowserWindowPlan().componentNamed (
		     TaskBrowserWindowHandler .SEARCH_DIRECTION_POPUP_NAME  );
    urlTextfield
      = (TextField)
         TaskBrowserWindowHandler.getTaskBrowserWindowPlan().componentNamed (
			       TaskBrowserWindowHandler . URL_TEXTFIELD_NAME );
    searchTextfield
      = (TextField)
         TaskBrowserWindowHandler.getTaskBrowserWindowPlan().componentNamed (
			    TaskBrowserWindowHandler . SEARCH_TEXTFIELD_NAME );
    htmlScrollgroup
      = (ScrollGroup)
         TaskBrowserWindowHandler.getTaskBrowserWindowPlan().componentNamed (
			   TaskBrowserWindowHandler .  HTML_SCROLLGROUP_NAME );

	/* Error-check */
    if (   ( backwardButton        == null )
	|| ( forwardButton         == null )
	|| ( reloadButton          == null )
	|| ( searchButton          == null )
	|| ( caseSensitiveButton   == null )
	|| ( searchDirectoryButton == null )
	|| ( internalExternalPopup == null )
	|| ( searchDirectionPopup  == null )
	|| ( urlTextfield          == null )
	|| ( searchTextfield       == null )
	|| ( htmlScrollgroup       == null ) )
    {
      System.err.println ( "[NewTaskWindowHandler:createAndInitialize]  "
	   + "Warning:  One or more Plan objects is NULL!!!\n"
	   + "    backwardButton........: " + backwardButton        + "\n"
	   + "    forwardButton.........: " + forwardButton         + "\n"
	   + "    reloadButton..........: " + reloadButton          + "\n"
	   + "    searchButton..........: " + searchButton          + "\n"
	   + "    caseSensitiveButton...: " + caseSensitiveButton   + "\n"
	   + "    searchDirectoryButton.: " + searchDirectoryButton + "\n"
	   + "    internalExternalPopup.: " + internalExternalPopup + "\n"
	   + "    searchDirectionPopup..: " + searchDirectionPopup  + "\n"
	   + "    urlTextfield..........: " + urlTextfield          + "\n"
	   + "    searchTextfield.......: " + searchTextfield       + "\n"
	   + "    htmlScrollgroup.......: " + htmlScrollgroup       + "\n"
	   + "------------------------------------------------------------" );
    }


	/* Reset the plan so that we can use it again later on... */
    TaskBrowserWindowHandler.getTaskBrowserWindowPlan() . releaseObjects();


	/* Perform object reconfigurations... */
    DefaultTextFilter.setDefaultTextFilter ( urlTextfield    );
    DefaultTextFilter.setDefaultTextFilter ( searchTextfield );
    urlTextfield    . setOwner ( this );
    searchTextfield . setOwner ( this );
    htmlTextview = new ExtendedTextView ( 0, 0,
			  htmlScrollgroup . scrollView() . bounds() . width,
			  htmlScrollgroup . scrollView() . bounds() . height );
    htmlScrollgroup . setContentView ( htmlTextview );
    htmlScrollgroup . setBackgroundColor ( htmlTextview . backgroundColor() );
    htmlTextview . setOwner ( this );
    htmlTextview . setEditable ( false );
    htmlTextview . setFont ( Font.fontNamed ( "Helvetica", Font.PLAIN, 18 ) );
    htmlTextview . setHorizResizeInstruction( View.WIDTH_CAN_CHANGE  );
    htmlTextview . setVertResizeInstruction ( View.HEIGHT_CAN_CHANGE );


	/* Report that we succeeded. */
    return true;
  }


  public RootView      getRootView() { return getBackwardButton().rootView(); }
  public Button           getBackwardButton()  { return backwardButton;       }
  public Button           getForwardButton()   { return forwardButton;        }
  public Button           getReloadButton()    { return reloadButton;         }
  public Button           getSearchButton()    { return searchButton;         }
  public Button           getCaseSensitiveButton()
					       { return caseSensitiveButton;  }
  public Button           getSearchDirectoryButton()
					       { return searchDirectoryButton;}
  public Popup            getInternalExternalPopup()
					       { return internalExternalPopup;}
  public Popup            getSearchDirectionPopup()
					       { return searchDirectionPopup; }
  public TextField        getUrlTextfield()    { return urlTextfield;         }
  public TextField        getSearchTextfield() { return searchTextfield;      }
  public ScrollGroup      getHtmlScrollgroup() { return htmlScrollgroup;      }
  public ExtendedTextView getHtmlTextview()    { return htmlTextview;         }

  public Vector           getURLVector()       { return urlVector;            }
  public int              getCurrentURLIndex() { return currentUrlIndex;      }


  public String getCurrentURL()
  {
    if ( getURLVector() . count() <= 0 )
      return null;

    return (String) (getURLVector() . elementAt ( getCurrentURLIndex() ));
  }

  public void decrementCurrentURL()
  {
    if ( getCurrentURLIndex() > 0 )
    {
      currentUrlIndex --;
      delayedReloadURL();
    }
  }

  public void incrementCurrentURL()
  {
    if ( getCurrentURLIndex() <  (getURLVector() . count() - 1) )
    {
      currentUrlIndex ++;
      delayedReloadURL();
    }
  }

  public void addURL ( String theURL )
  {
    File  urlFile;

    if (   ( theURL            != null )
	&& ( theURL . length()  > 0    )
	&& (   ( getCurrentURL() == null )
	    || ( getCurrentURL() . equals ( theURL ) == false ) ) )
    {
	/* Clean up URL's that don't start with 'file:' */
      if ( theURL . startsWith ( TaskBrowserWindowHandler.PATH_SEPARATOR ) )
	theURL = "file:" + theURL;


	/* IS this a file, and DOES this file exist? */
      if ( theURL . startsWith ( "file:" ) )
      {
	urlFile = new File ( theURL . substring ( 5 ) );
	if ( urlFile . exists() == false )
	{
	  runAlert ( "Error:",
		     "File \"" + theURL . substring ( 5 )
		     + "\" does not exist.",
		     "OK", null, null );
	  return;
	}
	if ( urlFile . canRead() == false )
	{
	  runAlert ( "Error:",
		     "Unable to read file \"" + theURL . substring ( 5 )
		     + "\".",
		     "OK", null, null );
	  return;
	}
      }

	/* Remove backtracked history */
      while ( getURLVector() . count()  >  (getCurrentURLIndex() + 1) )
      {
	getURLVector() . removeLastElement();
      }

	/* Truncate history, if necessary.  (Limiting us to MAX_HISTORY) */
      if ( getURLVector() . count() >= TaskBrowserWindowHandler.MAX_HISTORY )
	getURLVector() . removeFirstElement();

	/* Add theURL into the history... */
      getURLVector() . addElement ( theURL );

	/* Reset our current URL-history index */
      currentUrlIndex = getURLVector() . count() - 1;
    }

    delayedReloadURL();
  }
  
  public void delayedReloadURL()
  {
    getHtmlTextview() . overrideCursor ( View.WAIT_CURSOR );
    getHtmlTextview() . setString ( "Loading   " + getCurrentURL() );
    getHtmlTextview() . delayedRedraw();
    Application . application() . performCommandLater ( this,
			        TaskBrowserWindowHandler.REALLY_RELOAD_COMMAND,
			        null, true );
  }

  public void reloadURL()
  {
    if ( getCurrentURL() != null )
    {

      setTextFieldText ( getUrlTextfield(), getCurrentURL() );


      try 
      {
	getHtmlTextview() . importHTMLFromURL ( getCurrentURL() );
      }
      catch ( MalformedURLException theException )
      {
	runAlert ( "Error:",
		   "Bad URL: \"" + getCurrentURL() + "\"\n\n("
		   + theException.toString() + ")",
		   "OK", null, null );
	getHtmlTextview() . setString ( "" );
      }
      catch ( IOException theException )
      {
	runAlert ( "Error:",
		   "IO Exception while reading URL: \"" + getCurrentURL()
		   + "\"\n\n("
		   + theException.toString() + ")",
		   "OK", null, null );
	getHtmlTextview() . setString ( "" );
      }
      catch ( HTMLParsingException theException )
      {
	runAlert ( "Error:",
		   "HTML Parsing Error at line " + theException.lineNumber()
		   + ":\n\n" + theException.toString(),
		   "OK", null, null );
	getHtmlTextview() . setString ( "" );
      }


      getHtmlTextview() . resumeCursor();
      getHtmlTextview() . setFocusedView();
    }
  }


	/* synchronized to allow use of searchRange... */
  public synchronized void doSearch()
  {
    int           startPosition, endPosition, matchIndex;
    String        string;
    String        searchString  = getSearchTextfield().stringValue();
    Range         selectedRange = getHtmlTextview().selectedRange();
    final int     htmlTextViewLength = getHtmlTextview() . length();
    final boolean searchDirectionIsForward
		   = (   getSearchDirectionPopup() . selectedIndex()
		      == TaskBrowserWindowHandler.SEARCH_FORWARD_POPUP_INDEX );

	/* Find the text-range we are searching in */
    if ( searchDirectionIsForward )
    {
      if ( selectedRange . isNullRange() )
	startPosition = 0;

      else if ( selectedRange . length() <= 0 )
	startPosition = selectedRange . index() + 1;

      else
	startPosition = selectedRange . lastIndex() + 1;

	/* Idiocy check */
      if ( startPosition < 0 )
	startPosition = 0;
      if ( startPosition > htmlTextViewLength )
	startPosition = htmlTextViewLength;

      endPosition = htmlTextViewLength;
    }
    else /* Searching Backwards */
    {
      startPosition = 0;

      if ( selectedRange . isNullRange() )
	endPosition = htmlTextViewLength;

      else
	endPosition = selectedRange . index();

	/* Idiocy check */
      if ( endPosition < 0 )
	endPosition = 0;
      if ( endPosition > htmlTextViewLength )
	endPosition = htmlTextViewLength;
    }


	/* Obtain the corresponding string for that text-range */
    searchRange.index  = startPosition;
    searchRange.length = endPosition - startPosition;
    string = getHtmlTextview() . stringForRange ( searchRange );


	/* Are we doing a case-insensitive search? */
    if ( getCaseSensitiveButton() . state() == false )
    {
      searchString = searchString . toLowerCase();
      string       = string       . toLowerCase();
    }


	/* Find the index of the next matching string */
    if ( searchDirectionIsForward )
      matchIndex = string . indexOf ( searchString );

    else /* Searching Backwards */
      matchIndex = string . lastIndexOf ( searchString );


	/* Did we find a match? */
    if ( matchIndex == -1 )
    {
      if ( runAlert ( "Search:",
		 ( searchDirectionIsForward ? "End" : "Beginning" )
		 + " of Document reached.\nContinue from "
		 + ( searchDirectionIsForward ? "Beginning?" : "End?" ),
		 "Cancel", "Continue", null )
	   == Alert.SECOND_OPTION )
      {
	getHtmlTextview() . selectRange ( Range.nullRange() );
	Application . application()
		    . performCommandLater ( this,
				       TaskBrowserWindowHandler.SEARCH_COMMAND,
					    null );
      }
    }
    else /* match found */
    {
      searchRange . index  = startPosition + matchIndex;
      searchRange . length = searchString . length();
      getHtmlTextview() . setFocusedView();
      getHtmlTextview() . scrollRangeToVisible ( searchRange );
      getHtmlTextview() . selectRange          ( searchRange );
    }
  }



  public void performSearchDirectoryCommand()
  {
    if ( searchDirectoryWindowHandler == null )
    {
      searchDirectoryWindowHandler
	= new SearchDirectoryWindowHandler ( getIsExternalWindow(), this );

      searchDirectoryWindowHandler . runNoModal();

	/* Hack to display us properly,
	 * after Java finishes setting up the window...
	 */
      Application . application()
	. performCommandLater ( this,
			     TaskBrowserWindowHandler.SEARCH_DIRECTORY_COMMAND,
				null, false );
    }

    else if ( searchDirectoryWindowHandler . getWindow() . isVisible() )
    {
      searchDirectoryWindowHandler . getWindow() . moveToFront();
    }

    else
    {
      searchDirectoryWindowHandler . getWindow() . show();
      getVDT() . addWindow ( searchDirectoryWindowHandler . getWindow() );
    }


	/* Set the search-directory */
    String  currentUrl = getUrlTextfield() . stringValue();

    if ( currentUrl . startsWith ( "file:" ) )
    {
      String filepathSeparator
	      = System.getProperty ( "file.separator",
				     TaskBrowserWindowHandler.PATH_SEPARATOR );

      if ( TaskBrowserWindowHandler.PATH_SEPARATOR
	     . equals ( filepathSeparator )
	   == false )
      {
	if ( filepathSeparator . length() == 1 )
	{
	  searchDirectoryWindowHandler
	    . setSearchDirectory ( currentUrl . substring ( 5 ) . replace (
				       TaskBrowserWindowHandler.PATH_SEPARATOR
				         . charAt ( 0 ),
				       filepathSeparator . charAt ( 0 ) ) );
	}
	else
	{
	  System.err.println (
	      "[TaskBrowserWindowHandler:performSearchDirectoryCommand]  "
	      + "Warning:  Unable to tranlate URL to filename.  "
	      + "(filepathSeparator=\"" + filepathSeparator + "\")" );
	}
      }
      else /* Ie: PATH_SEPARATOR === filepathSeparator */
      {
	searchDirectoryWindowHandler
	  . setSearchDirectory ( currentUrl . substring ( 5 ) );
      }
    } /* if ( currentUrl . startsWith ( "file:" ) ) */

    else /* IF ( currentUrl . startsWith ( "file:" ) )  ... ELSE ... */
    {
      searchDirectoryWindowHandler
	. setSearchDirectory ( System.getProperty ( "user.dir" ) );
    }
  }





	/* ExtendedTarget Interface */
  public boolean canPerformCommand ( String theCommand ) 
  {
    return
         theCommand . equals ( TaskBrowserWindowHandler.BACKWARD_COMMAND     )
      || theCommand . equals ( TaskBrowserWindowHandler.FORWARD_COMMAND      )
      || theCommand . equals ( TaskBrowserWindowHandler.SEARCH_COMMAND       )
      || theCommand . equals ( TaskBrowserWindowHandler.RELOAD_COMMAND       )
      || theCommand . equals ( TaskBrowserWindowHandler.REALLY_RELOAD_COMMAND)
      || theCommand . equals ( TaskBrowserWindowHandler
			        . SEARCH_DIRECTORY_COMMAND );
  }


	/* Target Interface */
  public void performCommand ( String theCommand, Object theData )
  {
	/* Handle case of adding more arguments */
    if ( theCommand . equals ( TaskBrowserWindowHandler.BACKWARD_COMMAND ) )
    {
      decrementCurrentURL();
    }

    else if ( theCommand . equals ( TaskBrowserWindowHandler.FORWARD_COMMAND) )
    {
      incrementCurrentURL();
    }

    else if ( theCommand . equals ( TaskBrowserWindowHandler.SEARCH_COMMAND ) )
    {
      doSearch();
    }

    else if ( theCommand . equals ( TaskBrowserWindowHandler.RELOAD_COMMAND ) )
    {
      delayedReloadURL();
    }

    else if ( theCommand
	        . equals ( TaskBrowserWindowHandler.REALLY_RELOAD_COMMAND ) )
    {
      reloadURL();
    }

    else if ( theCommand
	        . equals ( TaskBrowserWindowHandler.SEARCH_DIRECTORY_COMMAND ))
    {
      performSearchDirectoryCommand();
    }

    else
    {
      super . performCommand ( theCommand, theData );
    }

  } /* public void performCommand ( String theCommand, Object theData ) */


	/* TextFieldOwner Interface */
  public void    textWasModified     ( TextField theTextField ) { }
  public void    textEditingDidBegin ( TextField theTextField ) { }
  public boolean textEditingWillEnd  ( TextField theTextField,
				       int       theEndCondition,
				       boolean   theContentsChanged )
							       { return true; }

  public void    textEditingDidEnd   ( TextField theTextField,
				       int       theEndCondition,
				       boolean   theContentsChanged )
  {
	/* Ignore any pending events after, or while, our window goes away. */
    if ( getWindow() . isVisible() == false )
      return;

    if ( theTextField == getUrlTextfield() )
    {
      String  urlString;
      String  basepath = getCurrentURL();

      if ( basepath == null )
	basepath = System.getProperty ( "user.dir" );

	/* Try to resolve relative paths... */
      urlString = WindowHandlerAbstractBaseClass.resolveFilepath (
				      getUrlTextfield() . stringValue(),
				      basepath,
				      null,
				      getRootView() );

	/* If we can't resolve the relative paths... */
      if ( urlString == null )
	urlString = getUrlTextfield() . stringValue();

      addURL ( urlString );
    }
  }


	/* TextViewOwner Interface */
  public  void attributesDidChange(TextView t, Range r){}
  public  void attributesWillChange(TextView t, Range r){}
  public  void textEditingDidBegin(TextView t) {}
  public  void textEditingDidEnd(TextView t) {}
  public  void selectionDidChange ( TextView theTextView )  {}
  public  void textWillChange ( TextView theTextView, Range theRange )  {}
  public  void textDidChange ( TextView theTextView, Range theRange )  {}

  public  void linkWasSelected ( TextView theTextView,
 				 Range    theRange,
				 String   theString )
  {
    String               tdlFilename;
    String               taskName;
    File                 newFile;
    DataTaskDefinition   dataTaskDefinition;
    String               openWindowCommand;
    Target               target;
    String               appendString;


	/* Is this a TDL link? */
    if (    theString . startsWith ( "file:/"                     )
	 && theString . endsWith   ( DataFile.TDL_FILENAME_SUFFIX ) )
    {
      tdlFilename = theString . substring ( 5 );
      taskName    = theTextView . stringForRange ( theRange );

	/* Do we already have this file loaded in? */
      if ( DataFile.getIsFilenameRegistered ( tdlFilename ) == false )
      {
	newFile = new File ( tdlFilename );

	  /* Does this file exist? */
	if ( newFile . exists() == false )
	{
	  runAlert ( "Error:",
		     "File \"" + tdlFilename + "\" does not exist.",
		     "OK", null, null );
	  return;
	}
	else /* file exists */
	{
	  try
	  {
	    DataFile . loadDataFile ( newFile );
	  }
	  catch ( DetailedParseException theException )
	  {
	    runAlert ( "Error:",
		       "Unable to parse File \"" + tdlFilename
		       + "\" as valid TDL code.\n\nParse Failure:\n"
		       + theException . toClippedString(),
		       "OK", null, null );
	    return;
	  }
	} /* if (file does not exist) ... else (file exists) ... */
      } /* if ( DataFile.getIsFilenameRegistered ( tdlFilename ) == false ) */


	/* Programmer error check... */
      if ( DataFile.getIsFilenameRegistered ( tdlFilename ) == false )
      {
	System.err.println ( "[TaskBrowserWindowHandler:linkWasSelected]  "
		     + "Programmer Error:  Failed to load/register file \""
		     + tdlFilename + "\"." );
      }


      dataTaskDefinition = DataTaskDefinition.getTaskForName ( taskName );
      if ( dataTaskDefinition != null )
      {
	if (    getInternalExternalPopup() . selectedIndex()
	     == TaskBrowserWindowHandler.INTERNAL_WINDOW_POPUP_INDEX )
	  openWindowCommand = VDT.NEW_INTERNAL_WINDOW;
	else
	  openWindowCommand = VDT.NEW_EXTERNAL_WINDOW;

	target = TargetChain . applicationChain()
			     . targetForCommand ( openWindowCommand );

	if ( target != null )
	{
	  Application . application()
		      . performCommandLater ( target, openWindowCommand,
					      dataTaskDefinition );
	}
	else
	{
	  System.err.println ( "[TaskBrowserWindowHandler:linkWasSelected]  "
	       + "Error: Unable to find TARGET for Opening New TDL Window. (\""
	       + openWindowCommand + "\")" );
	}
      }
      else  /* Ie:  dataTaskDefinition == null */
      {
	runAlert (
	    "Error:", "Task \"" + taskName
	    + "\" does not exist in any currently loaded files, including \""
	    + tdlFilename + "\".\n\nUnable to open Task...",
	    "OK", null, null );
	return;
      }

      return;

    } /* If ( this is a TDL link ) ... */




	/* Is this a full-link... */
    if (   theString. startsWith ( "file:"   )
	|| theString. startsWith ( "http:"   )
	|| theString. startsWith ( "ftp:"    )
	|| theString. startsWith ( "gopher:" ) )
    {
      addURL ( theString );
    }
    else  /* Relative link */
    {
      if (   ( getCurrentURL() != null )
	  && ( getCurrentURL()
	         . endsWith ( TaskBrowserWindowHandler.PATH_SEPARATOR ) ) )
      {
	appendString = "";
      }
      else
      {
	appendString = TaskBrowserWindowHandler.PATH_SEPARATOR;
      }

      addURL ( getCurrentURL() + appendString + theString );
    }
  }
}
