

import netscape.application.*;
import netscape.util.*;
import java.io.File;

public class VDT extends Application implements WindowOwner,
						Target, ExtendedTarget
{
	/* Global color'ings */
  public static Color WINDOW_BACKGROUND_COLOR     = Color.gray;
  public static Color PALETTE_BACKGROUND_COLOR    = Color.lightGray;
  public static Color SCROLLVIEW_BACKGROUND_COLOR
					  = GraphView.DEFAULT_BACKGROUND_COLOR;


	/* Constants */
  public final static int    DEFAULT_NEW_WINDOW_WIDTH   = 300;
  public final static int    DEFAULT_NEW_WINDOW_HEIGHT  = 200;
  public final static int    DEFAULT_MAIN_WINDOW_WIDTH  = 600;
  public final static int    DEFAULT_MAIN_WINDOW_HEIGHT = 500;
  public final static int    PALETTE_HEIGHT             = 30;
  public final static int    PALETTE_PAD                = 10;
  public final static int    PALETTE_OFFSET             = 5;
  public final static int    WINDOW_OFFSET              = 10;
  public final static int    MAX_WINDOW_OFFSET          = 100;


  public final static String COMMANDS_MENU         = "Commands";
  public final static String EDIT_MENU             = "Edit";
  public final static String TOOLS_MENU            = "Tools";
  public final static String WINDOWS_MENU          = "Windows";
  public final static String CLOSE_WINDOW_MENUITEM = "Close Window";

	/* Note: Cut, Copy, & Paste are in ExtendedTarget */
  public final static String QUIT                = "QUIT";
  public final static String CLOSE_WINDOW        = "CLOSE_WINDOW";
  public final static String DELETE         = DefaultTextFilter.DELETE_COMMAND;
  public final static String CREATE_NEW_TASK     = "CREATE_NEW_TASK";
  public final static String OPEN_TASK           = "OPEN_TASK";
  public final static String CREATE_HTML_DOCUMENTATION
					         = "CREATE_HTML_DOCUMENTATION";
  public final static String CREATE_TASK_BROWSER_INTERNAL
					         = "CREATE_BROWSER_INTERNAL";
  public final static String CREATE_TASK_BROWSER_EXTERNAL
					         = "CREATE_BROWSER_EXTERNAL";
  public final static String LOAD_FILE           = "LOAD_FILE";
  public final static String SAVE_TASKS_FILE     = "SAVE_CURRENT_TASKS_FILE";
  public final static String SAVE_FILE_AS        = "SAVE_FILE_AS";
  public final static String SAVE_ALL_FILES      = "SAVE_ALL_FILES";
  public final static String NEW_EXTERNAL_WINDOW = "NEW_EXTERNAL_WINDOW";
  public final static String NEW_INTERNAL_WINDOW = "NEW_INTERNAL_WINDOW";
  public final static String RAISE_WINDOW        = "RAISE_WINDOW";


	/* Instance Variables */
  protected ExternalWindow mainWindow;
  protected Vector         windows;
  protected ScrollGroup    palette;

  protected Menu           menu;
  protected MenuItem       menuCommands;
  protected Vector         menuWindows;
  protected ExtendedTarget previousView;

  protected int            internalWindowCreationCount;
  protected int            externalWindowCreationCount;
  protected int            menuDisableCount;

  protected FileChooser    loadFileChooser,
			   saveFileChooser;


	/* Instance Methods */
  public VDT ( )
  {
    super();

    mainWindow   = new ExternalWindow();

    setMainRootView ( mainWindow . rootView() );

    windows      = new Vector(1);
    palette      = null;
    menu         = new Menu();
    menuWindows  = new Vector(1); 
    previousView = null;

    internalWindowCreationCount = 0;
    externalWindowCreationCount = 0;
    menuDisableCount            = 0;

    loadFileChooser = null;
    saveFileChooser = null;

    createMenus ( menu, true, mainWindow );

    MenuItem  closeWindowMenuItem = findMenuItem ( menu,
						   VDT.COMMANDS_MENU,
						   VDT.CLOSE_WINDOW_MENUITEM );
    if ( closeWindowMenuItem != null )
      closeWindowMenuItem . setEnabled ( false );
    else
      System.err.println( "[VDT:VDT]  Warning:  closeWindowMenuItem == null" );

    mainWindow . setMenu ( menu );
    mainWindow . setResizable ( true );
    mainWindow . sizeTo ( VDT.DEFAULT_MAIN_WINDOW_WIDTH,
			  VDT.DEFAULT_MAIN_WINDOW_HEIGHT );
    mainWindow . setOwner ( this ); 
    mainWindow . rootView() . setColor ( VDT.WINDOW_BACKGROUND_COLOR );

    palette = createPalette ( mainWindow );
    mainWindow . addSubview ( palette );
  }

  public void init()
  {
	/* Initialize this now since it takes time & slows things down later */
    NewTaskWindowHandler.InitPlans();

	/* And lets make sure that we are a part of the Application-Chain... */
	/* (So other class-instances can easily get us to perform commands.) */
    TargetChain . applicationChain() . addTarget ( this, true );

	/* And now we can begin doing stuff... */
    mainWindow . show();
  }


  public FileChooser getLoadFileChooser()
  {
    if ( loadFileChooser == null )
    {
      loadFileChooser = new FileChooser ( mainWindow . rootView(),
					  "Load TDL File",
					  FileChooser.LOAD_TYPE );

      loadFileChooser . setDirectory( System.getProperty ( "user.dir", "/" ) );
    }

    return loadFileChooser;
  }

  public FileChooser getSaveFileChooser()
  {
    if ( saveFileChooser == null )
    {
      saveFileChooser = new FileChooser ( mainWindow . rootView(),
					  "Save To TDL File",
					  FileChooser.SAVE_TYPE );

      saveFileChooser . setDirectory( System.getProperty ( "user.dir", "/" ) );
    }

    return saveFileChooser;
  }

  public Vector         getWindows     ( ) { return windows;     }
  public ExternalWindow getMainWindow  ( ) { return mainWindow;  }
  public Menu           getMenu        ( ) { return menu;        }
  public Vector         getMenuWindows ( ) { return menuWindows; }


  protected MenuItem createMenus ( Menu            theMenu,
				   boolean         theSaveMenus,
				   ExternalWindow  theWindow )
  {
    MenuItem  commands, edit, tools, windows;

    commands = theMenu . addItemWithSubmenu ( VDT.COMMANDS_MENU );
    commands . submenu() . addItem ( "Create New Task", 'n',
				     VDT.CREATE_NEW_TASK, this )
                         . setData ( theWindow );
    commands . submenu() . addItem ( "Open Task", 'o',
				     VDT.OPEN_TASK, this )
                         . setData ( theWindow );
    commands . submenu() . addSeparator();
    commands . submenu() . addItem ( "Create HTML Documentation",
				     VDT.CREATE_HTML_DOCUMENTATION, this )
                         . setData ( theWindow );
    commands . submenu() . addItem ( "Task Browser (Internal)",
				     VDT.CREATE_TASK_BROWSER_INTERNAL, this )
                         . setData ( theWindow );
    commands . submenu() . addItem ( "Task Browser (External)", 't',
				     VDT.CREATE_TASK_BROWSER_EXTERNAL, this )
                         . setData ( theWindow );
    commands . submenu() . addSeparator();
    commands . submenu() . addItem ( "Load File", 'L',
				     VDT.LOAD_FILE, this )
                         . setData ( theWindow );
    commands . submenu() . addItem ( "Save Current Task's File",
				     VDT.SAVE_TASKS_FILE,
				     TargetChain.applicationChain() )
                         . setData ( theWindow );
//    commands . submenu() . addItem ( "Save File As",
//				     VDT.SAVE_FILE_AS, this );
    commands . submenu() . addItem ( "Save All Files",
				     VDT.SAVE_ALL_FILES, this )
                         . setData ( theWindow );
    commands . submenu() . addSeparator();
    commands . submenu() . addItem ( VDT.CLOSE_WINDOW_MENUITEM, 'w',
					 VDT.CLOSE_WINDOW, this )
                         . setData ( theWindow );
    commands . submenu() . addItem ( "Quit", 'q',
					 VDT.QUIT, this )
                         . setData ( theWindow );

    edit = theMenu . addItemWithSubmenu ( VDT.EDIT_MENU );
    edit . submenu() . addItem ( "Cut",    'x', ExtendedTarget.CUT,
				     TargetChain.applicationChain() );
    edit . submenu() . addItem ( "Copy",   'c', ExtendedTarget.COPY,
				     TargetChain.applicationChain() );
    edit . submenu() . addItem ( "Paste",  'v', ExtendedTarget.PASTE,
				     TargetChain.applicationChain() );
    edit . submenu() . addItem ( "Delete", 'd', VDT.DELETE,
				     TargetChain.applicationChain() );

    tools = theMenu . addItemWithSubmenu ( VDT.TOOLS_MENU );
    tools . submenu() . addItem ( "Color...",
				  ExtendedTarget.SHOW_COLOR_CHOOSER,
				  TargetChain.applicationChain() );
    tools . submenu() . addItem ( "Fonts...",
				  ExtendedTarget.SHOW_FONT_CHOOSER,
				  TargetChain.applicationChain() );
    

    windows = theMenu . addItemWithSubmenu ( VDT.WINDOWS_MENU );

    menuWindows . addElement ( windows );

    if ( theSaveMenus )
    {
      menuCommands = commands;
    }

    return windows;
  }


  public MenuItem findMenuItem ( Menu   theMenu,
				 String theMenuName,
				 String theMenuItemName )
  {
    int  i, j;

    for ( i=0;  i < theMenu.itemCount();  i++ )
    {
      if (   ( theMenu . itemAt ( i ) . title() . equals ( theMenuName ) )
	  && ( theMenu . itemAt ( i ) . hasSubmenu()                     ) )
      {
	for ( j=0;  j < theMenu . itemAt( i ) . submenu() . itemCount();  j++ )
	{
	  if ( theMenu . itemAt( i ) . submenu() . itemAt ( j ) . title()
	       . equals ( theMenuItemName ) )
	  {
	    return theMenu . itemAt( i ) . submenu() . itemAt ( j );
	  }
	}
      }
    }

    return null;
  }



  public boolean getIsMenusDisabled ( ) { return menuDisableCount > 0; }

  public void disableMenus ( )
  {
    Menu tmpMenu;
    int  i,j;

	/* If we are not currently disabled */
    if ( getIsMenusDisabled() == false )
    {
	/* Go through each menu on all the external windows */
	/* (Including the main window)                      */
      for ( i = -1;  i < getWindows().count();  i++ )
      {
	if ( i == -1 )
	  tmpMenu = getMenu();
	else if ( getWindows() . elementAt ( i ) instanceof ExternalWindow )
	  tmpMenu = ( (ExternalWindow) ( getWindows() . elementAt ( i ) ) )
	            . menu();
	else
	  continue;

	for ( j=0;  j < tmpMenu.itemCount();  j++ )
	  tmpMenu . itemAt ( j ) . setEnabled ( false );
      } /* For each window */
    }

    menuDisableCount++;
  }


  public boolean reenableMenus ( )
  {
    Menu tmpMenu;
    int  i,j;

	/* If we are currently disabled */
    if ( getIsMenusDisabled() == true )
    {
      menuDisableCount--;
	/* And we have just changed to being not-disabled */
      if ( getIsMenusDisabled() == false )
      {
	  /* Go through each menu on all the external windows */
	  /* (Including the main window)                      */
	for ( i = -1;  i < getWindows().count();  i++ )
	{
	  if ( i == -1 )
	    tmpMenu = getMenu();
	  else if ( getWindows() . elementAt ( i ) instanceof ExternalWindow )
	    tmpMenu = ( (ExternalWindow) ( getWindows() . elementAt ( i ) ) )
		      . menu();
	  else
	    continue;

	  for ( j=0;  j < tmpMenu.itemCount();  j++ )
	    tmpMenu . itemAt ( j ) . setEnabled ( true );
	} /* For each window */
      } /* if we have just changed to being not-disabled */
    } /* if we are  currently disabled */

    return getIsMenusDisabled();
  }


  protected void copyMenu ( Menu            theNewMenu,
			    ExternalWindow  theWindow  )
  {
    int   i;
    Menu  oldWindows = ( (MenuItem) getMenuWindows() . elementAt ( 0 ) )
		        . submenu(),
	  newWindows = createMenus( theNewMenu, false, theWindow ) . submenu();

    if ( oldWindows == null )
    {
      System.err.println ( "[VDT:copyMenu]  Warning:  "
			   + "No source windows menu." );
      return;
    }

    for ( i=0;  i < oldWindows.itemCount();  i++ )
    {
      newWindows . addItem ( oldWindows . itemAt ( i ) . title(),
			     VDT.RAISE_WINDOW, this )
	         . setData ( oldWindows . itemAt ( i ) . data() );
    }
  }


  protected ScrollGroup createPalette ( Window theWindow )
  {
    ScrollGroup   newPalette;

    int           i, x = VDT.PALETTE_PAD, y;
    View          container = new View ( 0, 0, 
					 VDT.DEFAULT_MAIN_WINDOW_WIDTH,
					 VDT.PALETTE_HEIGHT );
    PaletteView[]  paletteViews
		    = { new PaletteView ( "With...Do",
					  new DataWithDoStatement(),
					  DataWithDoStatement.WITH_DO_COLOR ),
			new PaletteView ( "Spawn", new DataSpawnTask(),
					  DataSpawnTask.SPAWN_COLOR    ),
			new PaletteView ( "C-Expr",
					  new DataExpressionStatement(),
					  DataExpressionStatement
					   .EXPRESSION_COLOR ),
			new PaletteView ( "C-Decl",
					  new DataDeclarationStatement(),
					  DataDeclarationStatement
					   .DECLARATION_COLOR ),
			new PaletteView ( "C-Jump",
					  new DataJumpStatement(),
					  DataJumpStatement.JUMP_COLOR ),
			new PaletteView ( "If-Else",
					  new DataIfElseStatement(),
					  DataIfElseStatement.IF_ELSE_COLOR),
			new PaletteView ( "Switch",
					  new DataSwitchStatement(),
					  DataSwitchStatement.SWITCH_COLOR ),
			new PaletteView ( "For",
					  new DataForStatement(),
					  DataForStatement.FOR_COLOR ),
			new PaletteView ( "While",
					  new DataWhileStatement(),
					  DataWhileStatement.WHILE_COLOR ),
			new PaletteView ( "Do-While",
					  new DataDoWhileStatement(),
					  DataDoWhileStatement.DO_WHILE_COLOR),
			new PaletteView ( "Compound-Statement",
					  new DataCompoundStatement(),
					  DataCompoundStatement
					   .COMPOUND_STATEMENT_COLOR )
		      };


    newPalette = new ScrollGroup ( );
    newPalette . setBorder ( new BezelBorder ( BezelBorder.GROOVED ) );
    newPalette . setVertScrollBarDisplay   ( ScrollGroup . NEVER_DISPLAY );
    newPalette . setHorizScrollBarDisplay  ( ScrollGroup . AS_NEEDED_DISPLAY );
    newPalette . setHorizResizeInstruction ( View.WIDTH_CAN_CHANGE  );
    newPalette . setContentView ( container );
    newPalette . scrollView() . setTransparent ( false );
    newPalette . setBackgroundColor ( VDT.PALETTE_BACKGROUND_COLOR );

    newPalette
      . setBounds ( VDT.PALETTE_OFFSET,
		    VDT.PALETTE_OFFSET, 
		    ( theWindow . contentSize() . width
		      - ( 2 * VDT.PALETTE_OFFSET ) ),
		    ( VDT.PALETTE_HEIGHT 
		      + newPalette . horizScrollBar() . minSize() . height )
		   );


    for ( i=0;  i < paletteViews.length; i++ )
    {
      container . addSubview ( paletteViews[i] );

      y = ( VDT.PALETTE_HEIGHT  -  paletteViews[i] . minSize() . height )  / 2;
      if ( y < 0 )
	y = 0;

      paletteViews[i] . moveTo ( x, y );

      x += paletteViews[i] . minSize() . width + VDT.PALETTE_PAD;

      if ( x > container . bounds() . width )
	container . sizeTo ( x, VDT.PALETTE_HEIGHT );
    }

	/* Shrink the container's size if we need to... */
    if ( x < container . bounds() . width )
      container . sizeTo ( x, VDT.PALETTE_HEIGHT );

    return newPalette;
  }




  public void addWindow ( Window theWindow )
  {
    int       i,j;

    getWindows() . addElement ( theWindow );

    for ( i=0;  i < getMenuWindows().count();  i++ )
    {
	/* Create a new menu item in each windows menu, */
	/* and attach the window to that menu item.     */
      ( (MenuItem) getMenuWindows() . elementAt ( i ) )
	 . submenu() . addItem ( theWindow . title(), VDT.RAISE_WINDOW, this )
	 . setData ( theWindow );
    }
  }
  public boolean removeWindow ( Window theWindow )
  {
    int     i, j;
    Menu    windowmenu;
    boolean returnValue;
    Vector  subviews, graphSubviews;

	/* Remove it from our list of windows... */
    returnValue = getWindows() . removeElementIdentical ( theWindow );

	/* Remove it from our sets of windows-menus... */
    for ( i=0;  i < getMenuWindows().count();  i++ )
    {
      windowmenu = ( (MenuItem) getMenuWindows().elementAt ( i ) ) . submenu();
      for ( j=0; j < windowmenu . itemCount(); j++ )
      {
	if ( windowmenu . itemAt ( j ) . data() == theWindow )
	{
	  windowmenu . removeItemAt ( j );
		/* Only one removal per menu */
	        /* Ie:  'j' is no longer an accurate index. */
	  break;
	}
      }
    }

	/* If we have a GraphView in this window, destroy it... */
    subviews = null;

    if ( theWindow instanceof InternalWindow )
    {
      subviews = ((InternalWindow) theWindow) . contentView() . subviews();
    }
    else if ( theWindow instanceof ExternalWindow )
    {
      subviews = ((ExternalWindow) theWindow) . rootView() . subviews();
    }

    for ( i=0;  (subviews != null)  &&  (i <  subviews . count());   i++ )
    {
      if (   ( subviews . elementAt ( i ) instanceof ScrollGroup )
	  && ( ( (ScrollGroup) ( subviews . elementAt ( i ) ) )
	       . contentView() instanceof GraphView              ) )
      {
	( (GraphView)
	  ( ((ScrollGroup) (subviews . elementAt ( i ))) . contentView() )
	 ) . getRootNode() . destroySubtree();
      }
    }


      /*Return based on whether or not we found the window in our master list*/
    return returnValue;
  }


  public void createWindow  ( boolean             theIsInternal,
			      DataTaskDefinition  theTask )
  {
    MenuItem        closeWindowMenuItem;
    Window          newWindow;
    int             minY, minSubviewY, windowCreationCount;
    GraphView       graphview;
    ScrollText      scrolltext;
    GraphNode       graphnode;
    GraphConnection graphconnection;
    ScrollGroup     scrollgroup;
    GraphModule     graphModule;

    if ( theIsInternal )
    {
      newWindow = new InternalWindow();
      minY = palette . bounds() . maxY() + VDT.PALETTE_OFFSET;
      minSubviewY = 0;
      windowCreationCount = internalWindowCreationCount;
      internalWindowCreationCount
	= (    ( internalWindowCreationCount + VDT.WINDOW_OFFSET )
	    %  VDT.MAX_WINDOW_OFFSET );


      ( (InternalWindow) newWindow ) . setCloseable ( true );
    }
    else
    {
      newWindow = new ExternalWindow();

	/* Configure the menus for the new external window */
      Menu      newMenu = new Menu();

      copyMenu ( newMenu, (ExternalWindow) newWindow );

      ( (ExternalWindow) newWindow ) . setMenu ( newMenu );


	/* Create a new palette for the new external window */
      ScrollGroup newPalette = createPalette ( newWindow );
      newWindow . addSubview ( newPalette );

	/* Compute our new minY's... */
      minY = menuCommands . minHeight();
      minSubviewY = newPalette . bounds() . maxY() + VDT.PALETTE_OFFSET;

      windowCreationCount = externalWindowCreationCount;
      externalWindowCreationCount
	= (    ( externalWindowCreationCount + VDT.WINDOW_OFFSET )
	    %  VDT.MAX_WINDOW_OFFSET );
    }

    newWindow . setTitle ( theTask.getTaskName() );
    newWindow . setResizable ( true );
    newWindow . sizeTo ( VDT.DEFAULT_NEW_WINDOW_WIDTH,
			 VDT.DEFAULT_NEW_WINDOW_HEIGHT );
    newWindow . setOwner ( this );

    graphview   = new GraphView();
    graphModule = theTask . createGraphModule ( GraphRoot.GUI_TREE_GRAPH );
    graphModule . insertModuleAt ( graphview . getRootNode() );


       /* Note: Adding Views seems to resize both that view & the subview.
	*       So Save & Restore the graphview's width & height!!!
	*/
    int width  = graphview . bounds() . width,
        height = graphview . bounds() . height;

    scrollgroup = new ScrollGroup ( 0, minSubviewY,
				    newWindow . contentSize() . width,
				    newWindow . contentSize() . height
				    - minSubviewY );
    scrollgroup . setHorizScrollBarDisplay ( ScrollGroup . AS_NEEDED_DISPLAY );
    scrollgroup . setVertScrollBarDisplay  ( ScrollGroup . AS_NEEDED_DISPLAY );
    scrollgroup . setContentView ( graphview );
    scrollgroup . setHorizResizeInstruction ( View.WIDTH_CAN_CHANGE  );
    scrollgroup . setVertResizeInstruction  ( View.HEIGHT_CAN_CHANGE );
    scrollgroup . scrollView() . setTransparent ( false );
//    scrollgroup . setBackgroundColor ( VDT.SCROLLVIEW_BACKGROUND_COLOR );
//    scrollgroup . scrollView() . setBuffered ( true );

    newWindow . addSubview ( scrollgroup );
    newWindow . moveTo ( windowCreationCount, minY + windowCreationCount );
    newWindow . show();

	/* Record this window */
    addWindow ( newWindow );


       /* Note: Adding Views seems to resize both that view & the subview.
	*       So Save & Restore the graphview's width & height!!!
	*/
    graphview . sizeTo ( width, height );
	/* Just to be safe -- there's a weird inconsistency bug here. */
    scrollgroup . scrollView() . subviewDidResize ( graphview );

	/* And lets try to make the graphview the window's focused view... */
    graphview . setFocusedView();
  }



  public boolean canPerformCommand ( String theCommand )
  {
    return
      (   ( theCommand . equals ( VDT.QUIT                         ) )
       || ( theCommand . equals ( VDT.CLOSE_WINDOW                 ) )
       || ( theCommand . equals ( VDT.CREATE_NEW_TASK              ) )
       || ( theCommand . equals ( VDT.OPEN_TASK                    ) )
       || ( theCommand . equals ( VDT.CREATE_HTML_DOCUMENTATION    ) )
       || ( theCommand . equals ( VDT.CREATE_TASK_BROWSER_INTERNAL ) )
       || ( theCommand . equals ( VDT.CREATE_TASK_BROWSER_EXTERNAL ) )
       || ( theCommand . equals ( VDT.LOAD_FILE                    ) )
       || ( theCommand . equals ( VDT.SAVE_TASKS_FILE              ) )
       || ( theCommand . equals ( VDT.SAVE_ALL_FILES               ) )
       || ( theCommand . equals ( VDT.NEW_EXTERNAL_WINDOW          ) )
       || ( theCommand . equals ( VDT.NEW_INTERNAL_WINDOW          ) )
       || ( theCommand . equals ( VDT.RAISE_WINDOW                 ) ) );
  }

  public void performCommand ( String theCommand, Object theData )
  {
    RootView   rootview = null;
    RootView   priorRootview;

    if (   ( theData != null                                          )
        && ( theData instanceof MenuItem                              )
	&& ( ((MenuItem) theData ) . data() != null                   )
	&& ( ((MenuItem) theData ) . data() instanceof ExternalWindow ) )
    {
      rootview = ((ExternalWindow) (((MenuItem) theData).data())) . rootView();
    }


    if ( theCommand . equals ( VDT.QUIT ) )
    {
      System.exit(0);
    }

    else if ( theCommand . equals ( VDT.CLOSE_WINDOW ) )
    {
      if (   ( theData instanceof MenuItem )
	  && ( ((MenuItem) theData) . data()  instanceof Window ) )
	( (Window)  ( (MenuItem) theData ) . data() ) . hide();
      else
	System.err.println ( "\n[VDT:performCommand]  WArning:  "
			     + "Unknown Window.  (" + theData + ")\n" );
    }

    else if ( theCommand . equals ( VDT.CREATE_NEW_TASK ) )
    {
      disableMenus();
      ( new NewTaskWindowHandler() ) . runModal ( rootview );
      reenableMenus();
    }

    else if ( theCommand . equals ( VDT.OPEN_TASK ) )
    {
      disableMenus();
      OpenTaskWindowHandler.getOpenTaskWindowHandler() . runModal ( rootview );
      reenableMenus();
    }

    else if ( theCommand . equals ( VDT.CREATE_HTML_DOCUMENTATION ) )
    {
      disableMenus();
      GenerateHTMLWindowHandler.getGenerateHTMLWindowHandler()
	. runModal ( rootview );
      reenableMenus();
    }

    else if ( theCommand . equals ( VDT.CREATE_TASK_BROWSER_INTERNAL ) )
    {
      ( new TaskBrowserWindowHandler ( false, this ) ) . run ( rootview );
    }

    else if ( theCommand . equals ( VDT.CREATE_TASK_BROWSER_EXTERNAL ) )
    {
      ( new TaskBrowserWindowHandler ( true, this  ) ) . run();
    }

    else if ( theCommand . equals ( VDT.LOAD_FILE ) )
    {
      if ( rootview == null )
	System.err.println ( "[VDT:performCommand(LOAD_FILE)]  Warning:  "
			     + "Programmer error:  No RootView refernece." );
      performLoadFileCommand ( rootview );
    }


    else if ( theCommand . equals ( VDT.SAVE_TASKS_FILE ) )
    {
      if ( rootview == null )
	System.err.println( "[VDT:performCommand(SAVE_TASKS_FILE)]  Warning:  "
			    + "Programmer error:  No RootView refernece." );

      boolean  noTaskToSave = true;

      if (   ( previousView != null         )
	  && ( previousView instanceof View ) )
      {
	View  view = (View) previousView;

	while (   (  view                            != null  )
	       && ( (view instanceof RootView      ) == false )
	       && ( (view instanceof InternalWindow) == false ) )
	{
	  view = view . superview();

	  if (   ( view instanceof ExtendedTarget )
	      && ( ( (ExtendedTarget) view )
		     . canPerformCommand ( VDT.SAVE_TASKS_FILE ) ) )
	  {
	    ((ExtendedTarget) view) . performCommand ( VDT.SAVE_TASKS_FILE,
						       theData );
	    noTaskToSave = false;
	    break; /* Exit WHILE loop */
	  }
	} /* WHILE ( ... ) */
      }

      if ( noTaskToSave )
      {
	UTIL.runAlert ( "Error:",
			"No Task Selected.  Unable to save file.",
			"OK", null, null, rootview );
      }
    }


    else if ( theCommand . equals ( VDT.SAVE_ALL_FILES ) )
    {
      if ( rootview == null )
	System.err.println ( "[VDT:performCommand(LOAD_FILE)]  Warning:  "
			     + "Programmer error:  No RootView refernece." );
      performSaveAllFilesCommand ( rootview );
    }

    else if (   ( theCommand .equals ( VDT.NEW_EXTERNAL_WINDOW ) )
	     || ( theCommand .equals ( VDT.NEW_INTERNAL_WINDOW ) ) )
    {
      if ( (theData instanceof DataTaskDefinition) == false )
      {
	System.err.println ( "[VDT:performCommand(new_*_Window)]  Error:  "
			     + "theData is not a DataTaskDefinition.  "
			     + "Aborting...." );
      }
      else
      {
	DataTaskDefinition  theTask = (DataTaskDefinition) theData;
	createWindow ( theCommand .equals ( VDT.NEW_INTERNAL_WINDOW ),
		       theTask );
      }
    }

    else if ( theCommand . equals ( VDT.RAISE_WINDOW ) )
    {
      if (   ( theData instanceof MenuItem )
	  && ( ((MenuItem) theData) . data()  instanceof Window ) )
	( (Window)  ( (MenuItem) theData ) . data() ) . moveToFront();
      else
	System.err.println ( "\n[VDT:performCommand]  Warning:  "
			     + "Unknown Window.  (" + theData + ")\n" );
    }

    else
    {
      System.err.println ( "\n[VDT:performCommand]  Warning:  "
			   + "Unknown command \"" + theCommand + "\"\n" );
    }
  }

  public void performLoadFileCommand ( RootView  theRootView )
  {
	/* Choose a file */
    getLoadFileChooser() . showModally();

	/* Try to load that file... */
    if ( getLoadFileChooser().file() != null )
    {
      File   fileToLoad = new File (   getLoadFileChooser().directory()
				     + getLoadFileChooser().file()      );

      if ( fileToLoad . exists() == false )
      {
	UTIL.runAlert ( 
	    "Error:", fileToLoad.getAbsolutePath() + "  Does Not Exist...",
	    "OK", null, null, theRootView );
      }

      else if ( fileToLoad . isDirectory() )
      {
	UTIL.runAlert (
	    "Error:",
	     fileToLoad.getAbsolutePath() + "  Is a Directory, NOT a File...",
	     "OK", null, null, theRootView );
      }

      else
      {
	try
	{
	  DataFile  dataFile = DataFile . loadDataFile ( fileToLoad );
	}
	catch ( DetailedParseException theException )
	{
	  UTIL.runAlert (
	      "Parse Error:",
	      "      In file: " + fileToLoad.getAbsolutePath() + "\n\n"
	      + theException . toClippedString(),
	      "OK", null, null, theRootView );
	}
      } /* else */
    } /* if ( user chose a file to load ) */
  }

  public void performSaveAllFilesCommand ( RootView theRootView )
  {
    Enumeration          knownFiles = DataFile.getRegisteredFiles();
    DataFile             dataFile;
    DataDestinationFile  dataDestinationFile = new DataDestinationFile();

    while ( knownFiles . hasMoreElements() )
    {
      dataFile = (DataFile) knownFiles.nextElement();

      if (    dataDestinationFile . setFile ( dataFile . getFilename() )
	   == false )
      {
	UTIL.runAlert (
	      "Error:",
	      "Unable to save file \"" + dataFile . getFilename() + "\".",
	      "OK", null, null, theRootView );
      }
      else
      {
	dataFile . generate ( dataDestinationFile,
			      DataComponent.ENTIRE_OBJECT );
	dataDestinationFile . close();
      }
    }
  }

	/* WindowOwner Interface */
    public void windowDidShow(Window theWindow) {}
    public boolean windowWillHide(Window theWindow)  {return true;}
    public boolean windowWillShow(Window theWindow)  {return true;}
    public void windowWillSizeBy(Window theWindow, Size theSize) {}
    public void windowDidBecomeMain ( Window theWindow ) { }

  public void windowDidResignMain ( Window theWindow )
  {
	/* Bug work-around.  Let menus in mainwindow control widgets in
	 * internal windows..
	 *
	 * (Effectively what happens is that the current window gets a
	 *  windowDidResignMain right before the menu is invoked.
	 *  Thus this hack...)
	 */
    if (   ( theWindow instanceof InternalWindow )
        && ( ( (InternalWindow) theWindow ) . focusedView() 
	     instanceof ExtendedTarget ) )
    {
      if ( previousView != null )
	TargetChain . applicationChain() . removeTarget ( previousView );
      previousView
	= (ExtendedTarget) ( ( (InternalWindow) theWindow ) . focusedView() );
      TargetChain . applicationChain() . addTarget ( previousView, true );
    }
  }


  public void windowDidHide ( Window theWindow )
  {
    if ( theWindow == getMainWindow() )
      System.exit(0);

    if ( removeWindow ( theWindow ) == false )
      System.err.println ( "[VDT:windowDidHide]  Warning:  Unknown window." );

	/* Bug work-around -- See windowDidResignMain */
    if (   ( theWindow instanceof InternalWindow           )
	&& (         previousView             != null      )
	&& ( ((View) previousView) . window() == theWindow ) )
    {
	/* Lose the reference to this window's focusedView()... */
      TargetChain . applicationChain() . removeTarget ( previousView );
      previousView = null;
    }
  }



  public static void main ( String args[] )
  {
    Parameters.setParameters ( args );
    VDT  theApplication = new VDT();
    theApplication . run();
    System . exit ( 0 );
  }
}
