
#ifndef stdtypes_H
#define stdtypes_H


typedef char                  int1;
typedef signed short int      int2;
typedef signed long  int      int4;
typedef signed long  long     int8;

typedef unsigned char         u_int1;
typedef unsigned short int    u_int2;
typedef unsigned long  int    u_int4;
typedef unsigned long  long   u_int8;


#ifdef NEEDS_BOOL
enum BOOLEAN
{
  FALSE = 0,
  TRUE  = (!FALSE)
};

#ifndef DISABLE_IOSTREAM
inline ostream &
operator << ( ostream & theOstream, const BOOLEAN & theBoolean )
{
  theOstream << ( ( theBoolean == TRUE ) ? "TRUE" : "FALSE" );
  return theOstream;
}
#endif /* DISABLE_IOSTREAM */

#endif  /* NEEDS_BOOL */


#ifndef HAS_STATUS
enum status_t
{
  FAILURE = FALSE,
  SUCCESS = TRUE
};

#ifndef DISABLE_IOSTREAM
inline ostream &
operator << ( ostream & theOstream, const status_t theStatus )
{
  theOstream << ( ( theStatus == SUCCESS ) ? "SUCCESS" : "FAILURE" );
  return theOstream;
}
#endif /* DISABLE_IOSTREAM */

#endif  /* ! HAS_STATUS */



#ifndef DISABLE_IOSTREAMBASE
class  iostreamBase
{
public:
  enum CONSTANTS
  {
    MAXIMUM_INDENT = 127
  };

  inline static void createSubIndentString (
		  char   theSubIndentString [ iostreamBase::MAXIMUM_INDENT ],
	    const char * theCurrentIndentString = ""                         );

  inline ostream & operator << ( ostream & theOstream ) const
				     { return printObject ( theOstream, "" ); }

  virtual ostream & printObject ( ostream    & theOstream,
				  const char * theIndentString = "") const = 0;
};

inline /*static*/ void
iostreamBase::createSubIndentString (
		  char   theSubIndentString [ iostreamBase::MAXIMUM_INDENT ],
	    const char * theCurrentIndentString = ""                         )
{
  strncpy ( theSubIndentString,
	    theCurrentIndentString,
	    iostreamBase::MAXIMUM_INDENT - 1 );
  theSubIndentString [ iostreamBase::MAXIMUM_INDENT - 1 ] = char(NULL);

  if ( strlen ( theSubIndentString ) <= (iostreamBase::MAXIMUM_INDENT - 3) )
    strcat ( theSubIndentString, "  " );

  else if (strlen ( theSubIndentString ) == (iostreamBase::MAXIMUM_INDENT - 2))
    strcat ( theSubIndentString, " " );
}

inline ostream &
operator << ( ostream & theOstream, const iostreamBase & theIostreamBase )
{
  return theIostreamBase . printObject ( theOstream, "" );
}
#endif /* ! DISABLE_IOSTREAMBASE */



  /*
   * Our TDL Standard Allocation Function signature...
   */
typedef  TCM_Task_Tree_Ref ( * _TDL_AllocationFunctionType ) ( const char * );



  /*
   * A nice, safe, double-checked strings equal routine.
   */
inline BOOLEAN StringEqual ( const char * theFirstString,
			     const char * theSecondString )
{
  if ( theFirstString == (const char *)NULL )
  {
    cerr << "[StringEqual]  Error:  First String is NULL." << endl;
    return FALSE;
  }

  else if ( theSecondString == (const char *)NULL )
  {
    cerr << "[StringEqual]  Error:  Second String is NULL." << endl;
    return FALSE;
  }
  
  else 
  {
    return strcmp ( theFirstString, theSecondString ) == 0 ? TRUE : FALSE;
  }
}


  /*
   * String duplication is sometimes necessary...
   * Note:  This returns heap-memory that must be delete'ed!!!!
   */
inline char * NewString ( const char * theStringToDuplicate )
{
  if ( theStringToDuplicate == (const char *)NULL )
  {
    cerr << "[NewString]  Error:  Original String is NULL!" << endl;
    return (char *)NULL;
  }
  else
  {
    size_t stringLength = strlen ( theStringToDuplicate ) + 1;
    char * newString    = new char [ stringLength ];

    strncpy ( newString, theStringToDuplicate, stringLength );

	/* Better play it safe and insure null-termination. */
    newString [ stringLength - 1 ] = char(NULL);

    return newString;
  }
}



  /* The nice thing about standards is that there are so many to choose from.*/
#if defined(sun) || defined(__sun) || defined (__sun__)

inline time_t
_TDL_MAKE_TIME_LOCAL ( struct tm * theTM ) { return timelocal ( theTM ); }
inline time_t
_TDL_MAKE_TIME_GMT   ( struct tm * theTM ) { return timegm    ( theTM ); }

#else /* defined ( sun || __sun || __sun__ ) */

inline time_t
_TDL_MAKE_TIME_LOCAL ( struct tm * theTM ) { return mktime    ( theTM ); }
inline time_t
_TDL_MAKE_TIME_GMT   ( struct tm * theTM ) { return mktime    ( theTM ); }

#endif /* defined ( sun || __sun || __sun__ ) */


	/* Useful routine to have around. */
inline struct timeval getTimeOfDay()
{
  struct timeval currentTime;
  if ( gettimeofday ( & currentTime, (struct timezone *) NULL ) != 0 )
  {
    cerr << "[getTimeOfDay]  Error:  gettimeofday() failed with errno="
	 << errno << endl;
    currentTime . tv_sec  = 0;
    currentTime . tv_usec = 0;
  }
  return currentTime;
}

	/* Useful routine to have around. */
inline struct timeval getTimeDifference ( struct timeval theFirstTime,
					  struct timeval theSecondTime )
{
  struct timeval  timeDifference, laterTime, earlierTime;

  if (   (     theFirstTime.tv_sec  >  theSecondTime.tv_sec    )
      || (   ( theFirstTime.tv_sec  == theSecondTime.tv_sec  )
	  && ( theFirstTime.tv_usec >= theSecondTime.tv_usec ) ) )
  {
    laterTime   = theFirstTime;
    earlierTime = theSecondTime;
  }
  else
  {
    laterTime   = theSecondTime;
    earlierTime = theFirstTime;
  }

  timeDifference.tv_sec  = laterTime.tv_sec  - earlierTime.tv_sec;
  timeDifference.tv_usec = laterTime.tv_usec - earlierTime.tv_usec;
  while ( timeDifference.tv_usec < 0 )
  {
    timeDifference.tv_sec  -= 1;
    timeDifference.tv_usec += 1000000;
  }

  return timeDifference;
}


  /* This is a nifty macro to have around...
   *   -- Watch out for side effects though!!!
   *      Ie: FLOATING_POINT_EQUIVALENT (i++,j++,1E-6) yields i+=2, j+=2...
   */
#define FLOATING_POINT_EQUIVALENT(X,Y,EPSILON)  \
                          ( ( ((X)>(Y)) ? ((X)-(Y)) : ((Y)-(X)) ) < (EPSILON) )

	/* Standard macros.   Beware side effects!! */
#define ABS(X)              ( ((X) >= 0) ? (X) : (- (X)) )
#define MIN(X,Y)            ( ((X) <= (Y)) ? (X) : (Y) )
#define MAX(X,Y)            ( ((X) >= (Y)) ? (X) : (Y) )
#define CLAMP(X,MINX,MAXX)  ( MIN( MAX( (X), (MINX) ), (MAXX) ) )
#define SQUARE(X)           ( (X) * (X) )


	/* Standard conversion macros... */
#define  RADIANS_TO_DEGREES(RADIANS)   ( double(RADIANS) * double(180) / M_PI )
#define  DEGREES_TO_RADIANS(DEGREES)   ( double(DEGREES) * M_PI / double(180) )


#endif /* stdtypes_H */

