
/*
 * This is the place for code that is utilized by compiled TDL code.
 * Naturally, this code can be used directly.  But its purpose is to
 * support (implement) specific TDL features.
 */

#ifndef TDL_H
#define TDL_H

 /* Standard C++ Header Files */
#include <iostream.h>
#include <iomanip.h>
#include <fstream.h>
#include <strstream.h>
#include <string.h>
#include <stdlib.h>
#include <math.h>
#include <sys/time.h>
#include <unistd.h>
#include <errno.h>

    /* TCM defines a const variable called MAXINT.
     * Unfortunately, so does math.h.  Only math.h #defines MAXINT...
     */
#undef MAXINT


 /* TCM Header Files */
#include <tcm.h>
#include <monitor.h>

 /* TDL Header Files */
#include "_TDL_BaseAction.H"
#include "stdtypes.H"
#include "tdldef.H"
#include "_TDL_InterfaceToTCM.H"


	/* Create this as a global, so that spawns outside of tasks can
	 * refer to "PARENT".  (Ie: For the #define to work right...)
	 * Now used in place of TCM_RootNode() to provide a rather hideous
	 * mechanism for inserting external nodes (spawns outside of tasks)
	 * into the TCM Task Tree at locations other than the TCM_RootNode.
	 */
extern TCM_Task_Tree_Ref const & _TDL_ENCLOSING_TASK; /* = TCM_RootNode() */

	/* Macro to mark variables as "USED" to the compiler, so as to   *
	 * prevent variable-not-used warnings on TDLC-created variables. *
	 * TDLC utilizes (generates) _TDL_MARKUSED(...) commands.        *
	 * If users utilize MARKUSED(...), the two can be distinguished. */
#define _TDL_MARKUSED(X)  ((void)(&(X)))
#define MARKUSED(X)       ((void)(&(X)))

	/* Monitor functions need to reference _TDL_ENCLOSING_TASK */
#define getNumberOfTriggers()    ( TCM_NumTriggers    ( _TDL_ENCLOSING_TASK ) )
#define getNumberOfActivates()   ( TCM_NumActivations ( _TDL_ENCLOSING_TASK ) )

	/* Note:  This is a case-insenstive macro.  The TDLC program         *
	 * automatically translates the keyword "TRIGGER" to all upper-case. */
#define TRIGGER()                ( TCM_Trigger ( _TDL_ENCLOSING_TASK ) )


	/* Trival means for user to access TDL Tasks as             *
	 * "const TCM_Task_Tree_Ref &"'s for use with TCM functions */
#define TDL_REF(X)   _TDL_SpawnedTasks [ # X ]


	/* Folks may want to implement an exception handler for BaseException*/
#define TDL_BaseException _TDL_BaseException

	/* Somewhat complicated way to access arguments from Tasks while
	 * outside the scope of those tasks...  Useful for POSTPONE'ed tasks.
	 * Ie:  TDL_TaskData_Ref          taskDataRef;
	 *      const TCM_Task_Tree_Ref & task;
         *      Goal foo ( int i ) { taskDataRef = TDL_TASKDATA();
         *                           task        = TDL_REF ( PARENT );
         *                           POSTPONE; }
         *      void bar() { cerr << _TDL_ARG(taskDataRef, foo, i);
         *                   TDL_SUCCESS ( task ); }
	 */
typedef TCM_Action_Ref TDL_TaskData_Ref;
typedef _Action        TDL_TaskData;
#define TDL_TASKDATA()    (this)
#define TDL_ARG( theTaskDataRef, theTaskName, theArgumentName )  \
 ( ((_TDL_ ## theTaskName *)((theTaskDataRef).operator*())) -> theArgumentName)

	/* POSTPONE'ed Tasks need a means to be marked SUCCEEDED or FAILED. */

	/* Ie: void bar() { TDL_SUCCESS ( task ); } */
#define TDL_SUCCESS( theTaskRefToMarkAsSucceeded )                            \
			   _TDL_DO_TCM_SUCCESS ( theTaskRefToMarkAsSucceeded )

	/* Note:  TheFailureArgs must be surrounded by parenthesis!!!!    */
	/* Ie: void bar() { TDL_FAIL ( task, engineOnFire, ("f3", 1) ); } */
#define TDL_FAIL( theTaskRefToFail, theFailureName, theFailureArgs )          \
  _TDL_DO_TCM_FAIL ( (theTaskRefToFail),                                      \
		     _TDL_Exception_ ## theFailureName ## ::exceptionMatches, \
		     _TDL_CreateException_ ## theFailureName theFailureArgs  )

	/* And POSTPONE'd Exception-Handlers need to be */ 
	/* able to access their Exception's data.       */
#define TDL_FAIL_DATA_ARG( theTaskDataRef, theExceptionName, theArgumentName )\
	      ( ( (_TDL_Exception_ ## theExceptionName *)                     \
		  ( ((theTaskDataRef).operator*()) -> getNodeRefCountData() ) \
		 ) -> theArgumentName )



	/* Macros to test if a TASK has completed execution. */
#define TDL_TASK_IS_COMPLETED( theTCMTaskTreeRef )                            \
	 ( TCM_IsPast ( TCM_EndOf ( TCM_AchievingOf ( theTCMTaskTreeRef ) ) ) )
#define TDL_IS_COMPLETED( theTask )                                           \
	 ( TDL_TASK_IS_COMPLETED ( TDL_REF ( theTask ) ) )

	/* Macros to test if a TASK is of a particular Task-type. */
#define TDL_TASK_INSTANCE_OF( theTCMTaskTreeRef, theTaskName )                \
	 ( StringEqual ( TCM_NodeName ( theTCMTaskTreeRef ), # theTaskName ) )
#define TDL_INSTANCE_OF( theTask, theTaskName )                               \
	 ( TDL_TASK_INSTANCE_OF ( TDL_REF ( theTask ), theTaskName ) )


	/* Macro to implement the real RESUME functionality. *
	 * Used internally to TDL.                           */
#define _TDL_INTERNAL_RESUME( theTaskClassName, theTCMTaskTreeRefToResume )   \
    ( (theTaskClassName *)                                                    \
      ( ( (_TDL_BaseAction *)                                                 \
	  ( TCM_GetAction ( theTCMTaskTreeRefToResume ) . operator*() )       \
	) -> setTcmTaskTreeNode ( theTCMTaskTreeRefToResume . operator *() )  \
      )                                                                       \
    )  -> resume

#define TDL_RESUME_ARGUMENT_DECLARATIONS  \
	  const TCM_Task_Tree_Ref & _TDL_ENCLOSING_TASK


#ifndef _TDL_INTERNAL_

#include "_TDL_Snode.H"
#include "_TDL_Slist.H"
#include "_TDL_Dnode.H"
#include "_TDL_Dlist.H"
#include "_TDL_ArrayList.H"
#include "_TDL_NamesList.H"
#include "_TDL_TreeNode.H"
#include "_TDL_TreeNodeBranch.H"
#include "_TDL_IterationIndex.H"
#include "_TDL_TDLStatement.H"
#include "_TDL_SpawnStatement.H"
#include "_TDL_SpawnStatementData.H"
#include "_TDL_SpawnStatementTreeNode.H"
#include "_TDL_Exceptions.H"
#include "_TDL_Constraints.H"
#include "_TDL_WithStatement.H"
#include "_TDL_WithStatementData.H"
#include "_TDL_WithStatementTreeNode.H"
#include "_TDL_HandleManager.H"
#include "_TDL_functions_2.H"

#endif /* ifndef _TDL_INTERNAL_ */



     /* Miscellaneous:
      *     Does our compiler conform to the new (1998) ANSI C++ Standard?
      *
      * Note:  The new [1998] ANSI C++ Standard has changed a number of things.
      * At present, both g++ 2.8.1 and egcs support the "new" Standard.  Which,
      * unfortunately, is syntactically incompatible with the "old" Standard
      * supported by g++ 2.7.2.x.
      */
#if    ( ! defined ( ANSI_STANDARD_CXX )                                  ) \
    && ( __GNUC__ == 2 && ( __GNUC_MINOR__ == 90 || __GNUC_MINOR__ == 8 ) )

#define ANSI_STANDARD_CXX 1

#endif 


#endif TDL_H

